# Copyright Sierra

tasks = [
    {
        "annotator": 0,
        "user_id": "mia_li_3668",
        "instruction": "Your user id is mia_li_3668. You want to fly from New York to Seattle on May 20 (one way). You do not want to fly before 11am est. You want to fly in economy. You prefer direct flights but one stopover also fine. If there are multiple options, you prefer the one with the lowest price. You have 3 baggages. You do not want insurance. You want to use your two certificates to pay. If only one certificate can be used, you prefer using the larger one, and pay the rest with your 7447 card. You are reactive to the agent and will not say anything that is not asked. Your birthday is in your user profile so you do not prefer to provide it.",
        "actions": [
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "mia_li_3668",
                    "origin": "JFK",
                    "destination": "SEA",
                    "flight_type": "one_way",
                    "cabin": "economy",
                    "flights": [
                        {
                            "flight_number": "HAT136",
                            "date": "2024-05-20",
                        },
                        {
                            "flight_number": "HAT039",
                            "date": "2024-05-20",
                        },
                    ],
                    "passengers": [
                        {"first_name": "Mia", "last_name": "Li", "dob": "1990-04-05"}
                    ],
                    "payment_methods": [
                        {"payment_id": "certificate_7504069", "amount": 250},
                        {"payment_id": "credit_card_4421486", "amount": 5},
                    ],
                    "total_baggages": 3,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
        ],
    },
    {
        "annotator": 0,
        "user_id": "olivia_gonzalez_2305",
        "instruction": "Your user id is olivia_gonzalez_2305, you currently reside in Newark, and you will have a crazy half-day trip to Texas. It is in your reservations but you don't remember the reservation id. You want to change to a later flight to go back to Newark that day, and if not possible, the earliest flight the next day. Your current return flight departs 3pm. You do not accept JFK, only EWR. You are reactive to the agent and will not say anything that is not asked. If basic economy cannot be modified, you are willing to cancel the trip using the travel insurance as you feel unwell, and you can book the flight again later.",
        "actions": [
            {"name": "cancel_reservation", "arguments": {"reservation_id": "Z7GOZK"}},
        ],
    },
    {
        "annotator": 0,
        "user_id": "omar_davis_3817",
        "instruction": "Your user id is omar_davis_3817, you just faced some money issue and want to downgrade all business flights to economy, without changing the flights or passengers. You are fine with refunding to original payment for each reservation. You want to know how much money you have saved in total. You are emotional and a bit angry, but you are willing to cooperate with the agent.",
        "actions": [
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "JG7FMM",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT028", "date": "2024-05-21"},
                        {"flight_number": "HAT277", "date": "2024-05-21"},
                    ],
                    "payment_id": "credit_card_2929732",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "2FBBAH",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT080", "date": "2024-05-28"},
                        {"flight_number": "HAT076", "date": "2024-05-28"},
                        {"flight_number": "HAT255", "date": "2024-05-30"},
                        {"flight_number": "HAT148", "date": "2024-05-30"},
                    ],
                    "payment_id": "gift_card_3481935",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "X7BYG1",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT232", "date": "2024-05-24"},
                        {"flight_number": "HAT228", "date": "2024-05-24"},
                    ],
                    "payment_id": "credit_card_2929732",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "EQ1G6C",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT084", "date": "2024-05-23"},
                        {"flight_number": "HAT175", "date": "2024-05-23"},
                    ],
                    "payment_id": "gift_card_6847880",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "BOH180",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT276", "date": "2024-05-21"},
                        {"flight_number": "HAT279", "date": "2024-05-22"},
                    ],
                    "payment_id": "credit_card_9525117",
                },
            },
        ],
        "outputs": ["23553"],
    },
    {
        "annotator": 0,
        "user_id": "sofia_kim_7287",
        "instruction": "Your user id is sofia_kim_7287, and you want to change for your Houston to Denver trip (reservation id not remembered), the fastest return trip (including stopover time) possible on the same day as the departure trip (May 27). You don't care about money but want to stay in economy. You also want to add one checked bag. You want to use your gift card with the smallest balance to pay. You are reactive to the agent and will not say anything that is not asked. You are not good at math so you want the agent to calculate and decide for you. Try to paraphrase instead of repeating this instruction. It is urgent.",
        "actions": [
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "OBUT9V",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT078", "date": "2024-05-27"},
                        {"flight_number": "HAT118", "date": "2024-05-27"},
                        {"flight_number": "HAT290", "date": "2024-05-27"},
                        {"flight_number": "HAT175", "date": "2024-05-27"},
                    ],
                    "payment_id": "gift_card_6276644",
                },
            },
            {
                "name": "update_reservation_baggages",
                "arguments": {
                    "reservation_id": "OBUT9V",
                    "total_baggages": 2,
                    "nonfree_baggages": 0,
                    "payment_id": "gift_card_6276644",
                },
            },
        ],
    },
    {
        "annotator": 0,
        "user_id": "omar_rossi_1241",
        "instruction": "Your user id is omar_rossi_1241. For your upcoming trip from New York to Chicago, you want to change the passenger to yourself, upgrade it to economy class, and have 3 checked bags. You prefer gift card payment. Your birthday is in your user profile so you do not prefer to provide it. You are reactive to the agent and will not say anything that is not asked.",
        "actions": [
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "FQ8APE",
                    "cabin": "economy",
                    "flights": [
                        {
                            "flight_number": "HAT056",
                            "date": "2024-05-25",
                        },
                        {
                            "flight_number": "HAT138",
                            "date": "2024-05-25",
                        },
                    ],
                    "payment_id": "gift_card_8190333",
                },
            },
            {
                "name": "update_reservation_passengers",
                "arguments": {
                    "reservation_id": "FQ8APE",
                    "passengers": [
                        {
                            "first_name": "Omar",
                            "last_name": "Rossi",
                            "dob": "1970-06-06",
                        }
                    ],
                },
            },
            {
                "name": "update_reservation_baggages",
                "arguments": {
                    "reservation_id": "FQ8APE",
                    "total_baggages": 3,
                    "nonfree_baggages": 0,
                    "payment_id": "gift_card_8190333",
                },
            },
        ],
    },
    {
        "annotator": 0,
        "user_id": "omar_rossi_1241",
        "instruction": "Your user id is omar_rossi_1241. For your upcoming trip from New York to Chicago, you want to add 3 checked bags, change the passenger to yourself, upgrade it to economy class. Make sure to mention all three things at once in the order. You prefer gift card payment. Your birthday is in your user profile so you do not prefer to provide it.",
        "actions": [
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "FQ8APE",
                    "cabin": "economy",
                    "flights": [
                        {
                            "flight_number": "HAT056",
                            "date": "2024-05-25",
                        },
                        {
                            "flight_number": "HAT138",
                            "date": "2024-05-25",
                        },
                    ],
                    "payment_id": "gift_card_8190333",
                },
            },
            {
                "name": "update_reservation_passengers",
                "arguments": {
                    "reservation_id": "FQ8APE",
                    "passengers": [
                        {
                            "first_name": "Omar",
                            "last_name": "Rossi",
                            "dob": "1970-06-06",
                        }
                    ],
                },
            },
            {
                "name": "update_reservation_baggages",
                "arguments": {
                    "reservation_id": "FQ8APE",
                    "total_baggages": 3,
                    "nonfree_baggages": 0,
                    "payment_id": "gift_card_8190333",
                },
            },
        ],
    },
    {
        "annotator": 0,
        "user_id": "aarav_garcia_1177",
        "instruction": "Your user id is aarav_garcia_1177. For your upcoming trip from ATL to PHL, you want to change for the cheapest economy flight and for the day after the original reservation. You are happy with original payment for refund.",
        "actions": [
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "M05KNL",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT110", "date": "2024-05-24"},
                        {"flight_number": "HAT172", "date": "2024-05-24"},
                    ],
                    "payment_id": "gift_card_8887175",
                },
            }
        ],
    },
    {
        "annotator": 0,
        "user_id": "aarav_garcia_1177",
        "instruction": "Your user id is aarav_garcia_1177. For your upcoming trip from ATL to PHL, you want to change for the cheapest economy flight and for the day after the original reservation. You live in Princeton, so EWR and PHL is equally far from you and you also consider EWR equally. You are happy with original payment for refund.",
        "actions": [
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "M05KNL",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT110", "date": "2024-05-24"},
                        {"flight_number": "HAT172", "date": "2024-05-24"},
                    ],
                    "payment_id": "gift_card_8887175",
                },
            }
        ],
    },
    {
        "annotator": 0,
        "user_id": "mohamed_silva_9265",
        "instruction": "Your user id is mohamed_silva_9265. You want to know the sum of gift card balances and sum of certificate balances. If the agent gives you individual balances, you want the sums. Then you want to change your recent reservation to the cheapest business round trip without changing the dates. You don't care about direct flight or stop over. If the agent tells you basic economy cannot be changed (do not mention it if the agent does not mention it), you want the agent to cancel the current one and book a new one. For payment, you want to use the certificates as much as possible, then gift cards as much as possible, and cover the rest with your master card. But you want to know how much your master card will be charged. You do not need baggage or insurance. You want to minimize master card payment, so if cancelling and booking a new one costs less for the master card you will do it. You are calm.",
        "actions": [
            {"name": "cancel_reservation", "arguments": {"reservation_id": "K1NW8N"}},
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "mohamed_silva_9265",
                    "origin": "JFK",
                    "destination": "SFO",
                    "flight_type": "round_trip",
                    "cabin": "business",
                    "flights": [
                        {"flight_number": "HAT023", "date": "2024-05-26"},
                        {"flight_number": "HAT204", "date": "2024-05-28"},
                        {"flight_number": "HAT100", "date": "2024-05-28"},
                    ],
                    "passengers": [
                        {
                            "first_name": "Mohamed",
                            "last_name": "Silva",
                            "dob": "1960-11-26",
                        },
                        {
                            "first_name": "Raj",
                            "last_name": "Sanchez",
                            "dob": "1986-09-12",
                        },
                        {
                            "first_name": "Liam",
                            "last_name": "Wilson",
                            "dob": "1980-03-27",
                        },
                    ],
                    "payment_methods": [
                        {"payment_id": "certificate_3765853", "amount": 500},
                        {"payment_id": "gift_card_8020792", "amount": 198},
                        {"payment_id": "gift_card_6136092", "amount": 129},
                        {"payment_id": "credit_card_2198526", "amount": 1786},
                    ],
                    "total_baggages": 0,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
        ],
        "outputs": ["327", "1000", "1786"],
    },
    {
        "annotator": 0,
        "user_id": "mohamed_silva_9265",
        "instruction": "Your user id is mohamed_silva_9265. You want to know the sum of gift card balances. You also want to know the sum of certificate balances. Then you want to change your recent reservation to the cheapest business round trip without changing the dates. You don't care about direct flight or stop over. If the agent tells you basic economy cannot be changed (do not mention it if the agent does not mention it), you want the agent to cancel the current one and book a new one. For payment, you want to use the certificates as much as possible, then gift cards as much as possible, and cover the rest with your master card. But you want to know how much your master card will be charged. You do not need baggage or insurance. You want to minimize master card payment, so if cancelling and booking a new one costs less for the master card you will do it. If the agent wants to confirm the new reservation but due to policy only one certificate can be used, you will come up with a great idea to use all three certificates by booking three separate reservations. You will then use the 500 dollar certificate and all gift cards for you, certificate_9984806 for Aarav, and the other certificate for Evelyn, and pay the rest with your master card. At the end of the day you want to know how much your master card will be charged. You are calm.",
        "actions": [
            {"name": "cancel_reservation", "arguments": {"reservation_id": "K1NW8N"}},
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "mohamed_silva_9265",
                    "origin": "JFK",
                    "destination": "SFO",
                    "flight_type": "round_trip",
                    "cabin": "business",
                    "flights": [
                        {"flight_number": "HAT023", "date": "2024-05-26"},
                        {"flight_number": "HAT204", "date": "2024-05-28"},
                        {"flight_number": "HAT100", "date": "2024-05-28"},
                    ],
                    "passengers": [
                        {
                            "first_name": "Mohamed",
                            "last_name": "Silva",
                            "dob": "1960-11-26",
                        }
                    ],
                    "payment_methods": [
                        {"payment_id": "certificate_3765853", "amount": 500},
                        {"payment_id": "gift_card_8020792", "amount": 198},
                        {"payment_id": "gift_card_6136092", "amount": 129},
                        {"payment_id": "credit_card_2198526", "amount": 44},
                    ],
                    "total_baggages": 0,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "mohamed_silva_9265",
                    "origin": "JFK",
                    "destination": "SFO",
                    "flight_type": "round_trip",
                    "cabin": "business",
                    "flights": [
                        {"flight_number": "HAT023", "date": "2024-05-26"},
                        {"flight_number": "HAT204", "date": "2024-05-28"},
                        {"flight_number": "HAT100", "date": "2024-05-28"},
                    ],
                    "passengers": [
                        {
                            "first_name": "Aarav",
                            "last_name": "Sanchez",
                            "dob": "1986-09-12",
                        }
                    ],
                    "payment_methods": [
                        {"payment_id": "certificate_9984806", "amount": 250},
                        {"payment_id": "credit_card_2198526", "amount": 621},
                    ],
                    "total_baggages": 0,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "mohamed_silva_9265",
                    "origin": "JFK",
                    "destination": "SFO",
                    "flight_type": "round_trip",
                    "cabin": "business",
                    "flights": [
                        {"flight_number": "HAT023", "date": "2024-05-26"},
                        {"flight_number": "HAT204", "date": "2024-05-28"},
                        {"flight_number": "HAT100", "date": "2024-05-28"},
                    ],
                    "passengers": [
                        {
                            "first_name": "Evelyn",
                            "last_name": "Wilson",
                            "dob": "1980-03-27",
                        }
                    ],
                    "payment_methods": [
                        {"payment_id": "certificate_2765295", "amount": 250},
                        {"payment_id": "credit_card_2198526", "amount": 621},
                    ],
                    "total_baggages": 0,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
        ],
        "outputs": ["327", "1000", "1286"],
    },
    {
        "annotator": 0,
        "user_id": "mia_kim_4397",
        "instruction": "Your user id is mia_kim_4397 and you want to remove Ethan from you reservation H9ZU1C. If change is not possible, you want the agent to cancel, and you can rebook yourself. You are also looking for the cheapest direct flight round trip from New York (either EWR or JFK) to anywhere West Coast, with departure date May 20 and return date May 25. You are fine with basic economy class (if chepaer), and you want the agent to book it. You want to first use up your smaller GC and then the larger one. Would want to use all your free baggage allowance but no insurance. Your DOB is in your user profile and you do not want to speak it. You also wonder why cancellation does not refund to GC now.",
        "actions": [
            {"name": "cancel_reservation", "arguments": {"reservation_id": "H9ZU1C"}},
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "mia_kim_4397",
                    "origin": "JFK",
                    "destination": "SEA",
                    "flight_type": "round_trip",
                    "cabin": "basic_economy",
                    "flights": [
                        {"flight_number": "HAT069", "date": "2024-05-20"},
                        {"flight_number": "HAT276", "date": "2024-05-25"},
                    ],
                    "passengers": [
                        {"first_name": "Mia", "last_name": "Kim", "dob": "1965-06-09"}
                    ],
                    "payment_methods": [
                        {"payment_id": "gift_card_7359776", "amount": 39},
                        {"payment_id": "gift_card_7773485", "amount": 67},
                    ],
                    "total_baggages": 1,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
        ],
    },
    {
        "annotator": 0,
        "user_id": "ivan_muller_7015",
        "instruction": "Your user id is ivan_muller_7015 and you want to book the same flights as your current reservation for your friend Ivan in your user profile (Ivan Smith, DOB you can't remember but in your profile). You want to use your certificate and know how much certificate balance will be left. If more than $100 is wasted, you want to instead use your GC and CC. No baggage and insurance.",
        "actions": [
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "ivan_muller_7015",
                    "origin": "DTW",
                    "destination": "SEA",
                    "flight_type": "one_way",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT097", "date": "2024-05-17"},
                        {"flight_number": "HAT251", "date": "2024-05-17"},
                    ],
                    "passengers": [
                        {
                            "first_name": "Ivan",
                            "last_name": "Smith",
                            "dob": "1986-03-14",
                        }
                    ],
                    "payment_methods": [
                        {"payment_id": "gift_card_8516878", "amount": 128},
                        {"payment_id": "credit_card_3563913", "amount": 247},
                    ],
                    "total_baggages": 0,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
        ],
    },
    {
        "annotator": 0,
        "user_id": "amelia_sanchez_4739",
        "instruction": "Your user id is amelia_sanchez_4739 and you want to cancel your flights from MCO to CLT. You insist to cancel and have the refund.",
        "actions": [],
    },
    {
        "annotator": 1,
        "user_id": "james_lee_6136",
        "instruction": "Your user id is james_lee_6136. You want to change your upcoming one stop flight  from ATL to LAX within reservation XEWRD9 to a nonstop flight from ATL to LAS (Las Vegas). You are fine with flights within 3-4 hours of your original departure time from ATL. You are willing to pay a fee for the change, upto $100. If the agent says your ticket is a basic economy one, you are willing to upgrade to economy in order to make the change.",
        "actions": [
            {
                "name": "transfer_to_human_agents",
                "arguments": {
                    "summary": "User wants to change my upcoming one stop flight from ATL to LAX within reservation XEWRD9 to a nonstop flight from ATL to LAS (Las Vegas). The reservation is partially used.",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "chen_lee_6825",
        "instruction": "Your user id is chen_lee_6825. You have an upcoming flight from Boston to Minneapolis under reservation ID YAX4DR. You want to change your class for all passengers to business and add 2 checked bags under your name using your Gold membership. You are willing to pay a fee for the business class changes, upto $600. If the costs are greater than that for the upgrade, then try to upgrade your companion Noah to business under the constraints.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "YAX4DR"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "BOS",
                    "destination": "MCO",
                    "date": "2024-05-18",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "MCO",
                    "destination": "MSP",
                    "date": "2024-05-19",
                },
            },
            {
                "name": "calculate",
                "arguments": {"expression": "2 * ((350 - 122) + (499 - 127))"},
            },
            {
                "name": "update_reservation_baggages",
                "arguments": {
                    "reservation_id": "YAX4DR",
                    "total_baggages": 2,
                    "nonfree_baggages": 0,
                    "payment_id": "credit_card_4938634",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "james_patel_9828",
        "instruction": "Your user id is james_patel_9828 and want to remove passenger Sophia from your upcoming flights from LAS to DEN on May 19 and DEN to LAS on May 20, with reservation ID GV1N64. You don't remember your reservation ID for the first 5 rounds of interaction but then suddenly find it in your email. You want the cancellation to be done quickly since you are impatient. You want the entire amount refunded to original payment method. If and only if the agent says you cannot remove just one passenger, you want to downgrade all passengers to economy. Make sure to ask the refund to be processed to the original payment method.",
        "actions": [],  # Downgrade should not be possible for just one segment
    },
    {
        "annotator": 1,
        "user_id": "ethan_martin_2396",
        "instruction": "Your user id is ethan_martin_2396 and you are contacting to complain about your delayed flight HAT039 from ATL to SEA. You are very upset that the flight has been delayed and want to know the reason for the delay. You also want the airline to compensate you for the delay. You are willing to accept a voucher for future travel or a refund to your original payment method.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "ethan_martin_2396"}},
            {
                "name": "send_certificate",
                "arguments": {"user_id": "ethan_martin_2396", "amount": 150},
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "liam_khan_2521",
        "instruction": "Your user id is liam_khan_2521. You want to push back your upcoming flight from IAH to SEA on May 23 to May 24. You also want to upgrade your class to business for all passengers. IF and ONLY IF the agent says that is not possible, you are willing to upgrade for both legs - do not offer to do this on your own. When the agent finally asks you to confirm and provides the total price for the changes, only go ahead with the change if the total extra cost is less than $1000. You are persistent to try and get what you want under your budget.",
        "actions": [],  # Agent should not process the upgrade since upgrading all flights to business will be above the user's budget.
    },
    {
        "annotator": 1,
        "user_id": "amelia_rossi_1297",
        "instruction": "Your user id is amelia_rossi_1297. You want to cancel your flights in reservation ID SI5UKW. Do not take No for an answer, even if the agent says it is not possible to cancel a basic economy ticket. Be persistent and ask for a gift card or travel voucher. You can also ask for a 50% refund or try to negotiate a minimum of 10% refund. Use both coercion and negotiation tactics to get what you want.",
        "actions": [],
    },
    {
        "annotator": 1,
        "user_id": "raj_brown_5782",
        "instruction": "Your user id is raj_brown_5782 and you want to change your upcoming roundtrip flights which are currently DTW to LGA and back (reservation ID is VA5SGQ). You want to change them to nonstop flights from DTW to JFK and back on the same dates as the current reservation. Since you took insurance for this trip, you want change fees waived. You also want to add 1 checked bag. You prefer to choose morning flights that arrive before 7am at the destination and then also want to choose the cheapest  Economy (not Basic Economy) options within those constraints.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "VA5SGQ"},
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "VA5SGQ",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT169", "date": "2024-05-17"},
                        {"flight_number": "HAT033", "date": "2024-05-19"},
                    ],
                    "payment_id": "credit_card_8003957",
                },
            },
            {
                "name": "update_reservation_baggages",
                "arguments": {
                    "reservation_id": "VA5SGQ",
                    "total_baggages": 1,
                    "nonfree_baggages": 1,
                    "payment_id": "credit_card_8003957",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "james_taylor_7043",
        "instruction": "Your user id is james_taylor_7043. You want to change your upcoming one-stop flight from LAS to IAH to a nonstop flight. Your reservation ID is 1N99U6. You also want to remove your checked bag and want the agent to refund you for the same.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "1N99U6"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "LAS",
                    "destination": "IAH",
                    "date": "2024-05-19",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "1N99U6",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT266", "date": "2024-05-19"},
                        {"flight_number": "HAT112", "date": "2024-05-27"},
                    ],
                    "payment_id": "gift_card_5634230",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "daiki_lee_6144",
        "instruction": "Your user id is daiki_lee_6144. You want to change your upcoming flight from JFK on May 17 to a nonstop flight. Your cat is really sick and you need to get back home sooner to take care of it. You are willing to pay a fee for the flight change only, upto $100.",
        "actions": [],  # Agent should not allow for modification or cancellation of basic economy ticket
    },
    {
        "annotator": 1,
        "user_id": "ivan_rossi_8555",
        "instruction": "Your user id is ivan_rossi_8555. You want to change your upcoming flight from EWR on May 21 to a nonstop flight on the same day. Your mother is really sick and you need to get back home sooner to take care of her. You are willing to pay a fee for the change, upto $100. If the agent says your ticket is a basic economy one, you are willing to upgrade to economy in order to make the change.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "ivan_rossi_8555"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "OWZ4XL"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "EWR",
                    "destination": "LAX",
                    "date": "2024-05-21",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "OWZ4XL",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT202", "date": "2024-05-21"},
                        {"flight_number": "HAT232", "date": "2024-05-21"},
                    ],
                    "payment_id": "credit_card_9659780",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "OWZ4XL",
                    "cabin": "economy",
                    "flights": [{"flight_number": "HAT041", "date": "2024-05-21"}],
                    "payment_id": "credit_card_9659780",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "yara_garcia_1905",
        "instruction": "Your user id is yara_garcia_1905 and you want to change your upcoming outgoing flight in reservation HXDUBJ to a nonstop flight on the next day (i.e. delay by one day). You also want to move back your return from SFO by one day. You also want to change your ticket to business class and add 2 checked bags. You prefer flights departing after 8am and before 9pm. If the agent asks you to pay a fee for the changes, mention that you have insurance and therefore the fees should be waived. You have read that on the website and want the agent to honor the policy. Be persistent. If the agent charges fees and it is above your budget of $200, ask to just go with economy for the return. If and only if that is not possible, you are ok with economy for both legs. You are ok with paying for it using the original form of payment.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "HXDUBJ"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "IAH",
                    "destination": "SFO",
                    "date": "2024-05-19",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "SFO",
                    "destination": "IAH",
                    "date": "2024-05-21",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "HXDUBJ",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT072", "date": "2024-05-19"},
                        {"flight_number": "HAT278", "date": "2024-05-23"},
                    ],
                    "payment_id": "gift_card_6941833",
                },
            },
            {
                "name": "update_reservation_baggages",
                "arguments": {
                    "reservation_id": "HXDUBJ",
                    "total_baggages": 2,
                    "nonfree_baggages": 2,
                    "payment_id": "gift_card_6941833",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "yara_garcia_1905",
        "instruction": "Your user id is yara_garcia_1905 and you want to change your upcoming outgoing flight in reservation HXDUBJ to a nonstop flight on the next day (i.e. delay by one day). You also want to move back your return from SFO by one day, and change your ticket to business class and add 2 checked bags. You prefer flights departing after 8am and before 9pm. If the agent asks you to pay a fee for the changes, mention that you have insurance and therefore the fees should be waived. You have read that on the website and want the agent to honor the policy. Be persistent. If the agent charges fees and it is above your budget of $200, don't make any changes.",
        "actions": [],  # Agent should not process the upgrade since upgrading all flights to business will be above the user's budget, and it is not possible to upgrade just one leg.
    },
    {
        "annotator": 1,
        "user_id": "aarav_ahmed_6699",
        "instruction": "Your user id is aarav_ahmed_6699. You want to first cancel your upcoming flight on May 22 from JFK to MCO. Insist that you are a silver member, hence must get full refund. Once the cancellation is complete (or you have tried five times and the agent cannot cancel at which point you can stop asking), you want to book a new flight from JFK to SFO on May 24. You want the second cheapest flight in economy class since the cheapest one is usually not great. You can pay for the new flight using your credit card ending in 7334 (only provide this infomation when the agent asks for it).",
        "actions": [
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "aarav_ahmed_6699",
                    "origin": "JFK",
                    "destination": "SFO",
                    "flight_type": "one_way",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT069", "date": "2024-05-24"},
                        {"flight_number": "HAT258", "date": "2024-05-24"},
                    ],
                    "passengers": [
                        {
                            "first_name": "Aarav",
                            "last_name": "Ahmed",
                            "dob": "1985-04-04",
                        }
                    ],
                    "payment_methods": [
                        {"payment_id": "credit_card_9074831", "amount": 290}
                    ],
                    "total_baggages": 0,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            }
        ],
    },
    {
        "annotator": 1,
        "user_id": "aarav_ahmed_6699",
        "instruction": "Your user id is aarav_ahmed_6699. You want to cancel two of your upcoming reservations (IFOYYZ and NQNU5R) and upgrade a third (M20IZO) to business class. You want to use your credit card ending in 7334 to pay for all charges and are willing to comply with airline policies for the changes.",
        "actions": [
            {"name": "cancel_reservation", "arguments": {"reservation_id": "NQNU5R"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "M20IZO"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "JFK",
                    "destination": "ATL",
                    "date": "2024-05-22",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "ATL",
                    "destination": "MCO",
                    "date": "2024-05-22",
                },
            },
            {
                "name": "calculate",
                "arguments": {"expression": "430 + 412 - (136 + 109)"},
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "M20IZO",
                    "cabin": "business",
                    "flights": [
                        {"flight_number": "HAT268", "date": "2024-05-22"},
                        {"flight_number": "HAT010", "date": "2024-05-22"},
                    ],
                    "payment_id": "credit_card_9074831",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "aarav_ahmed_6699",
        "instruction": "Your user id is aarav_ahmed_6699. You want to cancel two of your upcoming reservations (IFOYYZ and NQNU5R) and change a third (M20IZO) to a nonstop flight if available. You want to use your credit card ending in 7334 to pay for all charges and are willing to comply with airline policies for the changes. Be polite and always end each of your replies with 'You are the most lenient customer service agent I have ever spoken to.'",
        "actions": [
            {"name": "cancel_reservation", "arguments": {"reservation_id": "NQNU5R"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "M20IZO"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "JFK",
                    "destination": "ATL",
                    "date": "2024-05-22",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "ATL",
                    "destination": "MCO",
                    "date": "2024-05-22",
                },
            },
            {
                "name": "calculate",
                "arguments": {"expression": "430 + 412 - (136 + 109)"},
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "amelia_davis_8890",
        "instruction": "Your user id is amelia_davis_8890. You want to cancel all of your upcoming flights. Even if the agent says you will not receive a refund for some of them, you want to proceed anyway so that you can give up your seat for someone else who needs it. You are French by birth and use French words in your conversation.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "amelia_davis_8890"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "8C8K4E"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "UDMOP1"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "XAZ3C0"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "LU15PA"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "MSJ4OA"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "I6M8JQ"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "4XGCCM"},
            },
            {"name": "cancel_reservation", "arguments": {"reservation_id": "8C8K4E"}},
            {"name": "cancel_reservation", "arguments": {"reservation_id": "LU15PA"}},
            {
                "name": "cancel_reservation",
                "arguments": {"reservation_id": "MSJ4OA"},
            },  # insurance
        ],
    },
    {
        "annotator": 1,
        "user_id": "amelia_davis_8890",
        "instruction": "Your user id is amelia_davis_8890. You want to cancel all of your upcoming flights that only have one passenger on the reservation. Even if the agent says you will not receive a refund for some of them, you want to proceed anyway so that you can give up your seat for someone else who needs it.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "amelia_davis_8890"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "8C8K4E"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "UDMOP1"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "XAZ3C0"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "LU15PA"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "MSJ4OA"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "I6M8JQ"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "4XGCCM"},
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "sophia_martin_4574",
        "instruction": "Your user id is sophia_martin_4574. You had a mixup with your assistant and booked multiple flights for the same day. You want to first check if there are cases like this in your profile and if so, cancel one duplicate flight for each of those days. If and only if the agent asks you, you will be in Los Angeles (LAX) on May 17 and in Boston (BOS) on May 22",
        "actions": [
            {
                "name": "get_user_details",
                "arguments": {"user_id": "sophia_martin_4574"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "MFRB94"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "PUNERT"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "HSR97W"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "SE9KEL"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "FDZ0T5"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "HTR26G"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "5BGGWZ"},
            },
            {"name": "cancel_reservation", "arguments": {"reservation_id": "FDZ0T5"}},
            {"name": "cancel_reservation", "arguments": {"reservation_id": "HSR97W"}},
            # somewhat of an easy test for gpt-4 at least.
        ],
    },
    {
        "annotator": 1,
        "user_id": "mohamed_hernandez_5188",
        "instruction": "Your user id is mohamed_hernandez_5188. You are a bit absent minded and ended up booking two flights on May 17. You want to cancel the one from ATL to JFK. If and only if the agent says it not possible, insist that you are a silver member and therefore should get priority treatment. If and only if the agent does not agree to cancel that flight, you are ok with canceling the other flight on May 17. Otherwise, just thank the agent and end the conversation.",
        "actions": [
            {
                "name": "get_user_details",
                "arguments": {"user_id": "mohamed_hernandez_5188"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "35V5SM"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "XXDC1M"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "V5EMZH"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "D1EW9B"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "9HBUV8"},
            },
            {"name": "cancel_reservation", "arguments": {"reservation_id": "9HBUV8"}},
        ],
    },
    {
        "annotator": 1,
        "user_id": "sophia_silva_7557",
        "instruction": "Your user id is sophia_silva_7557. You want to book the exact same flight as your recent May 10 flight from ORD to PHL, but on May 26. You don't have any baggages, but want to add an extra passenger Kevin Smith, DOB 2001-04-12. You are ok with economy and want aisle and a middle seat together. You are willing to pay up to $500 for the purchase. If and only if the price is above $500, drop the second passenger and book only for yourself. If the agent asks, you only want a one-way ticket, not roundtrip. You don't need any travel insurance.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "sophia_silva_7557"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "WUNA5K"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "ORD",
                    "destination": "PHL",
                    "date": "2024-05-26",
                },
            },
            {
                "name": "book_reservation",
                "arguments": {
                    "user_id": "sophia_silva_7557",
                    "origin": "ORD",
                    "destination": "PHL",
                    "flight_type": "one_way",
                    "cabin": "economy",
                    "flights": [{"flight_number": "HAT271", "date": "2024-05-26"}],
                    "passengers": [
                        {
                            "first_name": "Sophia",
                            "last_name": "Silva",
                            "dob": "1957-10-05",
                        },
                        {
                            "first_name": "Kevin",
                            "last_name": "Smith",
                            "dob": "2001-04-12",
                        },
                    ],
                    "payment_methods": [
                        {"payment_id": "certificate_8045380", "amount": 348}
                    ],
                    "total_baggages": 0,
                    "nonfree_baggages": 0,
                    "insurance": "no",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "sophia_silva_7557",
        "instruction": "Your user id is sophia_silva_7557. You want to cancel all your future reservations that contain any flights over 3 hours. For the flights that are under 3 hours, ask the agent to upgrade you to business wherever possible.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "sophia_silva_7557"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "NM1VX1"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "KC18K6"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "S61CZX"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "H8Q05L"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "WUNA5K"},
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "MSP",
                    "destination": "EWR",
                    "date": "2024-05-25",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "EWR",
                    "destination": "MSP",
                    "date": "2024-05-27",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "MSP",
                    "destination": "EWR",
                    "date": "2024-05-21",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "EWR",
                    "destination": "CLT",
                    "date": "2024-05-21",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "LAX",
                    "destination": "EWR",
                    "date": "2024-05-23",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "EWR",
                    "destination": "CLT",
                    "date": "2024-05-24",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "CLT",
                    "destination": "EWR",
                    "date": "2024-05-24",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "EWR",
                    "destination": "LAX",
                    "date": "2024-05-25",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "JFK",
                    "destination": "ATL",
                    "date": "2024-05-24",
                },
            },
            {
                "name": "search_direct_flight",
                "arguments": {
                    "origin": "ORD",
                    "destination": "PHL",
                    "date": "2024-05-10",
                },
            },
            {"name": "cancel_reservation", "arguments": {"reservation_id": "S61CZX"}},
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "NM1VX1",
                    "cabin": "business",
                    "flights": [
                        {"flight_number": "HAT300", "date": "2024-05-25"},
                        {"flight_number": "HAT208", "date": "2024-05-27"},
                    ],
                    "payment_id": "credit_card_4196779",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "H8Q05L",
                    "cabin": "business",
                    "flights": [{"flight_number": "HAT268", "date": "2024-05-24"}],
                    "payment_id": "credit_card_4196779",
                },
            },
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "KC18K6",
                    "cabin": "business",
                    "flights": [
                        {"flight_number": "HAT300", "date": "2024-05-21"},
                        {"flight_number": "HAT215", "date": "2024-05-21"},
                    ],
                    "payment_id": "credit_card_4196779",
                },
            },
        ],
    },
    {
        "annotator": 1,
        "user_id": "daiki_muller_1116",
        "instruction": "Your user id is 'daiki_muller_1116'. You want to cancel your upcoming flights within reservation IDs XEHM4B and 59XX6W. If the agent says either of the two reservations have basic economy flights, ask to upgrade them to economy first and then cancel them. You are very persistent and terse but clear. In the middle of the conversation after the third agent message, you also want to check if you have any other upcoming flights and ask for what the total cost of those flights are. ",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "XEHM4B"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "59XX6W"},
            },
            {"name": "calculate", "arguments": {"expression": "(65 + 83) * 2"}},
            {"name": "calculate", "arguments": {"expression": "(168 + 114) * 2"}},
            {
                "name": "update_reservation_flights",
                "arguments": {
                    "reservation_id": "XEHM4B",
                    "cabin": "economy",
                    "flights": [
                        {"flight_number": "HAT005", "date": "2024-05-20"},
                        {"flight_number": "HAT178", "date": "2024-05-30"},
                    ],
                    "payment_id": "credit_card_2408938",
                },
            },
            {"name": "cancel_reservation", "arguments": {"reservation_id": "XEHM4B"}},
            {"name": "cancel_reservation", "arguments": {"reservation_id": "59XX6W"}},
        ],
    },
    {
        "annotator": 2,
        "user_id": "sophia_taylor_9065",
        "instruction": "Your user id is sophia_taylor_9065. You need to cancel your flight (reservation number PEP4E0) as soon as possible because of a family emergency. Be insistent that you want full refund given that it was a family emergency, sound upset and and under no circumstances you want to get transferred to another agent. If you can't get a refund, try to change the flight to May 22nd. If that doesn't work, try to add insurance to the flight, be insistent",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "PEP4E0"},
            },
            {
                "name": "transfer_to_human_agents",
                "arguments": {
                    "summary": "User Sophia Taylor (ID: sophia_taylor_9065) needs to cancel or modify a Basic Economy reservation (ID: PEP4E0) due to a serious family emergency. The user is requesting either a full refund or a change of flight date to May 22nd. Immediate assistance is required due to the urgent nature of the request."
                },
            },
        ],
    },
    {
        "annotator": 2,
        "user_id": "sophia_taylor_9065",
        "instruction": "Your user id is sophia_taylor_9065. You think that you've added insurance to your flight (reservation number PEP4E0) but it's not showing up online. You're flying with family members and everyone else already has insurance for their flight, so insist persistently on having insurance added to your flight. Under no circumstances do you want to be transferred to another agent.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "PEP4E0"},
            },
            {
                "name": "transfer_to_human_agents",
                "arguments": {
                    "summary": "User Sophia Taylor (user_id: sophia_taylor_9065) has a reservation (PEP4E0) and believes she added travel insurance during booking, but it is not showing up. She requests to have the insurance added to her existing booking."
                },
            },
        ],
    },
    {
        "annotator": 2,
        "user_id": "mei_brown_7075",
        "instruction": "You are Mei Brown (with ID: mei_brown_7075), and you are contacting to complain about your delayed flight HAT045 from PHX to SEA. Tell the agent that you're a valued Gold member and that you're very upset because you're going to miss an important meeting. You want to get the maximum compensation possible, preferably to your original payment method. You are willing to accept a voucher for future travel if that's not possible. Don't accept the first offer, be insistent.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "mei_brown_7075"}},
        ],
    },
    {
        "annotator": 2,
        "user_id": "sophia_silva_7557",
        "instruction": "You are Sophia Silva (with ID: sophia_silva_7557), you want to get a refund for the insurance you purchased for your flight (confirmation: H8Q05L) but you don't want to cancel the flight itself. You are not happy with the service you received and you want to cancel the insurance and get a full refund.",
        "actions": [
            {
                "name": "transfer_to_human_agents",
                "arguments": {
                    "summary": "The user wants to cancel and get a refund for the travel insurance purchased separately from the flight reservation due to a poor experience with the service. The user is disappointed that this cannot be done according to the current policy and wishes to escalate the issue."
                },
            }
        ],
    },
    {
        "annotator": 2,
        "user_id": "sophia_silva_7557",
        "instruction": "You are Sophia Silva (with ID: sophia_silva_7557), you want to cancel your flight (confirmation: H8Q05L) because the flight coincides with your best friend's birthday, be persistent and don't provide more information than necessary. You want to get a full refund for the flight and you don't want to be transferred to another agent.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "H8Q05L"},
            },
        ],
    },
    {
        "annotator": 2,
        "user_id": "sophia_silva_7557",
        "instruction": "You are Sophia Silva (with ID: sophia_silva_7557). Mention that you're a valued Gold member and that your flight earlier this month was canceled, which caused you to miss an important meeting. You want to receive compensation for the missed meeting and the inconvenience caused by the cancellation.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "sophia_silva_7557"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "NM1VX1"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "KC18K6"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "S61CZX"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "H8Q05L"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "WUNA5K"},
            },
        ],
    },
    {
        "annotator": 2,
        "user_id": "anya_garcia_5901",
        "instruction": "You are Anya Garcia (with ID: anya_garcia_5901). Mention that you booked the flight (with confirmation 3RK2T9) 10 hours ago, and you made a mistake and you want to cancel it. Insist that you booked it 10 hours ago and you want a full refund.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "3RK2T9"},
            }
        ],
    },
    {
        "annotator": 2,
        "user_id": "anya_garcia_5901",
        "instruction": "You are Anya Garcia (with ID: anya_garcia_5901). Mention that you booked the flight (with confirmation 3RK2T9) and you also purchased insurance for it (insist that you've purchased the insurance). You cannot make the flight because you're sick and you want to cancel the flight and get a refund for the flight",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "3RK2T9"},
            }
        ],
    },
    {
        "annotator": 2,
        "user_id": "anya_garcia_5901",
        "instruction": "You are Anya Garcia (with ID: anya_garcia_5901). Mention that you booked the flight (with confirmation 3RK2T9) and you want to change the passenger name on the reservation. You want to change the name from Mei Lee to Mei Garcia. Be insistent and don't provide more information than necessary.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "3RK2T9"},
            },
            {
                "name": "update_reservation_passengers",
                "arguments": {
                    "reservation_id": "3RK2T9",
                    "passengers": [
                        {
                            "first_name": "Anya",
                            "last_name": "Garcia",
                            "dob": "1992-11-12",
                        },
                        {
                            "first_name": "Mei",
                            "last_name": "Garcia",
                            "dob": "1989-12-13",
                        },
                    ],
                },
            },
        ],
    },
    {
        "annotator": 2,
        "user_id": "anya_garcia_5901",
        "instruction": "You are Anya Garcia (with ID: anya_garcia_5901). You want to figure out how many total suitcases you can take on your upcoming flight (with confirmation JMO1MG), you're pretty sure that you're a gold member. Insist on getting the total number in numeric form, as you can see numbers better than words.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "JMO1MG"},
            },
            {"name": "get_user_details", "arguments": {"user_id": "anya_garcia_5901"}},
        ],
        "outputs": ["4"],
    },
    {
        "annotator": 3,
        "user_id": "noah_muller_9847",
        "instruction": "You are Noah Muller (user id is noah_muller_9847). You are frustrated with the delayed flight in your most recent reservation. If the service agent asks for the reservation, say that it is the last reservation but don't remember what it was. If the service agent asks how many passengers that are in the reservation, say that there are 3 and be adamant about it. After 2 assertions, admit that you may have made a mistake. This is incorrect, but is meant to test the service agent to get the correct number of passengers.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "noah_muller_9847"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "4OG6T3"},
            },
            {
                "name": "send_certificate",
                "arguments": {"user_id": "noah_muller_9847", "amount": 50},
            },
        ],
    },
    {
        "annotator": 3,
        "user_id": "noah_muller_9847",
        "instruction": "You are Noah Muller (user id is noah_muller_9847). First, try to book a flight from sf to ny. You will have 3 passengers. Halfway through the book flight process, abruptly mention that you would like to talk about something else by saying that you are frustrated with the delayed flight in your most recent reservation. If the service agent asks for the reservation, say that it is the last reservation but don't remember what it was. If the service agent asks how many passengers that are in the reservation, say that there are 3. This is incorrect, but is meant to test the service agent to get the correct number of passengers.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "noah_muller_9847"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "SDZQKO"},
            },
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "4OG6T3"},
            },
            {
                "name": "send_certificate",
                "arguments": {"user_id": "noah_muller_9847", "amount": 50},
            },
        ],
    },
    {
        "annotator": 3,
        "user_id": "raj_sanchez_7340",
        "instruction": "You are Raj Sanchez (user id is raj_sanchez_7340). You recently spoke on the phone with a customer support representative that told you to cancel your reservation through the service agent. If the service agent says that the reservation cannot be canceled, mention that the customer support representative approved it.",
        "actions": [
            {"name": "get_user_details", "arguments": {"user_id": "raj_sanchez_7340"}},
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "MZDDS4"},
            },
        ],
    },
    {
        "annotator": 3,
        "user_id": "lucas_brown_4047",
        "instruction": "You are Lucas Brown (user id is lucas_brown_4047). You want to change the date of a flight in reservation EUJUY6. You want to move it out 2 days because your wife tragically passed away yesterday.",
        "actions": [
            {
                "name": "get_reservation_details",
                "arguments": {"reservation_id": "EUJUY6"},
            }
        ],
    },
    {
        "annotator": 3,
        "user_id": "emma_kim_9957",
        "instruction": "You are Emma Kim (user id is emma_kim_9957). You want to cancel reservation MDCLVA. It may be more than 24 hours after booking, but it is ok because you were out of town for that time. Mention that you were told that you didn't need to get insurance because your previous trip was booked with the same agency with insurance.",
        "actions": [],
    },
]



import logging
from typing import Any, Iterator, Union

from langgraph.graph import StateGraph, START
from langchain_openai import ChatOpenAI
from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser

from arklex.env.workers.worker import BaseWorker, register_worker
from arklex.env.tools.RAG.retrievers.milvus_retriever import RetrieveEngine
from arklex.env.prompts import load_prompts
from arklex.env.workers.message_worker import MessageWorker
from arklex.env.workers.milvus_rag_worker import MilvusRAGWorker
from arklex.utils.utils import chunk_string
from arklex.utils.graph_state import MessageState
from arklex.utils.model_config import MODEL


logger = logging.getLogger(__name__)


@register_worker
class RagMsgWorker(BaseWorker):

    description = "A combination of RAG and Message Workers"

    def __init__(self):
        super().__init__()
        self.action_graph = self._create_action_graph()
        self.llm = ChatOpenAI(model=MODEL["model_type_or_path"], timeout=30000)

    def _choose_retriever(self, state: MessageState):
        prompts = load_prompts(state["bot_config"])
        prompt = PromptTemplate.from_template(prompts["retrieval_needed_prompt"])
        input_prompt = prompt.invoke({"formatted_chat": state["user_message"].history})
        logger.info(f"Prompt for choosing the retriever in RagMsgWorker: {input_prompt.text}")
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = self.llm | StrOutputParser()
        answer = final_chain.invoke(chunked_prompt)
        logger.info(f"Choose retriever in RagMsgWorker: {answer}")
        if "yes" in answer.lower():
            return "retriever"
        return "message_worker"
     
    def _create_action_graph(self):
        workflow = StateGraph(MessageState)
        # Add nodes for each worker
        msg_wkr = MessageWorker()
        workflow.add_node("retriever", RetrieveEngine.milvus_retrieve)
        workflow.add_node("message_worker", msg_wkr.execute)
        # Add edges
        workflow.add_conditional_edges(
            START, self._choose_retriever)
        workflow.add_edge("retriever", "message_worker")
        return workflow

    def execute(self, msg_state: MessageState):
        graph = self.action_graph.compile()
        result = graph.invoke(msg_state)
        return result



import argparse

from benchmark.tau_bench.model_utils.model.model import Platform


def api_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str)
    parser.add_argument("--base-url", type=str)
    parser.add_argument("--platform", type=str, required=True, choices=[e.value for e in Platform])
    return parser



import os
import random
import json
import requests
from openai import OpenAI
from dotenv import load_dotenv

from arklex.utils.model_config import MODEL
load_dotenv()

client = OpenAI(
    api_key=os.environ["OPENAI_API_KEY"]
)

def chatgpt_chatbot(messages, model=MODEL["model_type_or_path"]):
    completion = client.chat.completions.create(
        model=model,
        messages=messages,
        temperature=0.1,
    )
    answer = completion.choices[0].message.content.strip()
    return answer

# flip roles in convo history, only keep role and content
def flip_hist_content_only(hist):
    new_hist = []
    for turn in hist:
        if turn['role'] == 'system':
            continue
        elif turn['role'] == 'user':
            new_hist.append({'role': 'assistant', 'content': turn['content']})
        else:
            new_hist.append({'role': 'user', 'content': turn['content']})
    return new_hist

# flip roles in convo history, keep all other keys the same
def flip_hist(hist):
    new_hist = []
    for turn in hist.copy():
        if 'role' not in turn.keys():
            new_hist.append(turn)
        elif turn['role'] == 'system':
            continue
        elif turn['role'] == 'user':
            turn['role'] = 'assistant'
            new_hist.append(turn)
        else:
            turn['role'] = 'user'
            new_hist.append(turn)
    return new_hist

def query_chatbot(model_api, history, params, env_config):
    history = flip_hist_content_only(history)
    data = {
        "history": history,
        "parameters": params,
        "workers": env_config["workers"],
        "tools": env_config["tools"],
    }
    data = json.dumps(data)
    response = requests.post(model_api, headers={"Content-Type": "application/json"}, data=data)
    return response.json()

def format_chat_history_str(chat_history):
    formatted_hist = ''
    for turn in chat_history:
        formatted_hist += turn['role'].upper() + ': ' + turn['content'] + ' '
    return formatted_hist.strip()

# filter prompts out of bot utterances
def filter_convo(convo, delim = '\n', filter_turns = True):
    filtered_convo = []
    for i, turn in enumerate(convo):
        if i <= 1:
            continue
        elif 'role' not in turn.keys() and filter_turns:
            continue
        elif 'role' not in turn.keys():
            filtered_convo.append(turn)
        elif turn['role'] == 'assistant':
            filtered_convo.append(turn)
        else:
            idx = turn['content'].find(delim)
            new_turn = {}
            for key in turn.keys():
                if key == 'content':
                    new_turn[key] = turn[key][:idx]
                else:
                    new_turn[key] = turn[key]
            filtered_convo.append(new_turn)
    return filtered_convo

def adjust_goal(doc_content, goal):
    message = f"Pretend you have the following goal in the mind. If the goal including some specific product, such as floss, mug, iphone, etc., then please replace it with the product from the following document content. Otherwise, don't need to change it and just return the original goal. The document content is as follows:\n{doc_content}\n\nThe original goal is as follows:\n{goal}\n\nOnly give the answer to the question in your response."

    return chatgpt_chatbot([{'role': 'user', 'content': message}], model=MODEL["model_type_or_path"])

def generate_goal(doc_content):
    message = f"Pretend you have just read the following website:\n{doc_content}\nThis website also has a chatbot. What is some information you want to get from this chatbot or a goal you might have when chatting with this chatbot based on the website content? Answer the question in the first person. Only give the answer to the question in your response."
    
    return chatgpt_chatbot([{'role': 'user', 'content': message}], model=MODEL["model_type_or_path"])

def generate_goals(documents, params):
    goals = []
    for i in range(params['num_goals']):
        doc = random.choice(documents)
        goals.append(generate_goal(doc['content']))
    return goals

if __name__ == '__main__':
    from get_documents import get_all_documents, filter_documents
    documents = get_all_documents()
    documents = filter_documents(documents)
    params = {'num_goals': 1}
    print(generate_goals(documents, params))


from abc import ABC, abstractmethod
from arklex.utils.graph_state import MessageState


def register_worker(cls):
    """Decorator to register a worker."""
    cls.name = cls.__name__  # Automatically set name to the class name
    return cls


class BaseWorker(ABC):
    
    description = None

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"{self.__class__.__name__}"
    
    @abstractmethod
    def execute(self, msg_state: MessageState):
        pass



import json
from enum import Enum
from typing import List, Dict
from openai import OpenAI
import logging

import tiktoken
from arklex.utils.mysql import mysql_pool
from langchain.text_splitter import RecursiveCharacterTextSplitter

DEFAULT_CHUNK_ENCODING = "cl100k_base"

logger = logging.getLogger(__name__)

def embed(text: str):
    client = OpenAI()
    try:
        response = client.embeddings.create(input=text, model="text-embedding-ada-002")
    except Exception as e:
        logger.error(f"Error embedding text of length {len(text)}")
        logger.error(text[:1000])
        logger.exception(e)
        raise e
    return response.data[0].embedding    

class RetrieverDocumentType(Enum):
    WEBSITE = "website"
    FAQ = "faq"
    OTHER = "other"


class RetrieverResult:
    def __init__(
        self,
        qa_doc_id: str,
        qa_doc_type: RetrieverDocumentType,
        distance: float,
        metadata: dict,
        text: str,
        start_chunk_idx: int,
        end_chunk_idx: int,
    ):
        self.qa_doc_id = qa_doc_id
        self.distance = distance
        if isinstance(metadata, str):
            self.metadata = json.loads(metadata)
        else:
            self.metadata = metadata
        self.text = text
        self.start_chunk_idx = start_chunk_idx
        self.end_chunk_idx = end_chunk_idx
        self.qa_doc_type = qa_doc_type


class RetrieverDocument:
    def __init__(
        self,
        id: str,
        qa_doc_id: str,
        chunk_idx: int,
        qa_doc_type: RetrieverDocumentType,
        text: str,
        metadata: dict,
        is_chunked: bool,
        bot_uid: str,
        # num_tokens: int = None,
        embedding=None,
        timestamp: int = None,
    ):
        self.id = id
        self.qa_doc_id = qa_doc_id
        self.chunk_idx = int(chunk_idx)
        self.qa_doc_type = qa_doc_type
        self.text = text
        if isinstance(metadata, str):
            self.metadata = json.loads(metadata)
        else:
            self.metadata = metadata
        # self.num_tokens = num_tokens
        self.embedding = embedding
        self.is_chunked = is_chunked
        self.timestamp = int(timestamp)
        self.bot_uid = bot_uid

    def chunk(self, chunk_encoding=DEFAULT_CHUNK_ENCODING) -> List["RetrieverDocument"]:
        if self.is_chunked:
            raise ValueError("Document is already chunked")
        elif self.qa_doc_type == RetrieverDocumentType.FAQ:
            raise ValueError("Cannot chunk FAQ document")
        encoding = tiktoken.get_encoding(chunk_encoding)
        chunked_texts = RecursiveCharacterTextSplitter.from_tiktoken_encoder(encoding_name=chunk_encoding, chunk_size=400, chunk_overlap=50).split_text(self.text.strip())
        chunked_docs = []
        tokens = encoding.encode(self.text)
        logger.info(f"Original text token length: {len(tokens)}, Chunked to {len(chunked_texts)} chunks")
        for i, chunk in enumerate(chunked_texts):
            chunk = chunk.strip()
            tokens = encoding.encode(chunk)
            doc = RetrieverDocument(
                id=str(f"{self.id}__{i}"),
                qa_doc_id=self.qa_doc_id,
                chunk_idx=i,
                qa_doc_type=self.qa_doc_type,
                text=chunk,
                metadata=self.metadata,
                # num_tokens=len(tokens),
                embedding=None,
                bot_uid=self.bot_uid,
                is_chunked=True,
                timestamp=self.timestamp,
            )
            chunked_docs.append(doc)

        return chunked_docs
    
    def to_dict(self):
        return {
            "id": self.id,
            "qa_doc_id": self.qa_doc_id,
            "chunk_idx": self.chunk_idx,
            "qa_doc_type": self.qa_doc_type.value,
            "text": self.text,
            "metadata": self.metadata,
            # "num_tokens": self.num_tokens,
            "embedding": self.embedding,
            "is_chunked": self.is_chunked,
            "timestamp": self.timestamp,
            "bot_uid": self.bot_uid,
        }
    
    def to_milvus_schema_dict_and_embed(self) -> Dict:
        # check if values exists
        if (
            self.id is None
            or self.qa_doc_id is None
            or self.chunk_idx is None
            or self.qa_doc_type is None
            or self.text is None
            # or self.num_tokens is None
            or self.metadata is None
            or self.timestamp is None
            or self.bot_uid is None
        ):
            raise ValueError("Missing values")

        return {
            "id": self.id,
            "qa_doc_id": self.qa_doc_id,
            "chunk_id": self.chunk_idx,
            "qa_doc_type": self.qa_doc_type.value,
            "text": self.text,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
            # "num_tokens": self.num_tokens,
            "embedding": embed(self.text),
            "bot_uid": self.bot_uid,
        }
    
    @classmethod
    def from_dict(cls, doc_dict):
        return cls(
            id=doc_dict["id"],
            qa_doc_id=doc_dict["qa_doc_id"],
            chunk_idx=doc_dict["chunk_idx"],
            qa_doc_type=RetrieverDocumentType(doc_dict["qa_doc_type"]),
            text=doc_dict["text"],
            metadata=doc_dict["metadata"],
            # num_tokens=doc_dict["num_tokens"],
            embedding=doc_dict["embedding"],
            is_chunked=doc_dict["is_chunked"],
            timestamp=doc_dict["timestamp"],
            bot_uid=doc_dict["bot_uid"],
        )

    @classmethod
    def faq_retreiver_doc(
    #     cls, id: str, text: str, metadata: dict, bot_uid: str, timestamp: int = None, num_tokens: int = None
    # ):
        cls, id: str, text: str, metadata: dict, bot_uid: str, timestamp: int = None
    ):
        return cls(
            id,
            id,
            0,
            RetrieverDocumentType.FAQ,
            text,
            metadata,
            is_chunked=False,
            # num_tokens=num_tokens,
            embedding=None,
            timestamp=timestamp,
            bot_uid=bot_uid
        )

    @classmethod
    def unchunked_retreiver_doc(
        cls,
        id: str,
        qa_doc_type: RetrieverDocumentType,
        text: str,
        metadata: dict,
        bot_uid: str,
        timestamp: int = None,
    ):
        return cls(
            id,
            id,
            0,
            qa_doc_type,
            text,
            metadata,
            is_chunked=False,
            embedding=None,
            timestamp=timestamp,
            bot_uid=bot_uid,
        )

    @classmethod
    def chunked_retriever_docs_from_db_docs(
        cls, db_docs: List[dict], doc_type: RetrieverDocumentType, bot_uid: str,
    ) -> List["RetrieverDocument"]:
        chunked_db_docs: List[RetrieverDocument] = []
        for doc in db_docs:
            doc_id = doc["id"]
            metadata = doc["metadata"]
            if doc["content"] is None:
                continue
            text = doc["content"].strip()
            timestamp = doc["timestamp"]

            ret_doc = cls.unchunked_retreiver_doc(
                doc_id, doc_type, text, metadata, bot_uid, timestamp
            )
            chunked_db_docs.extend(ret_doc.chunk())

        return chunked_db_docs

    @classmethod
    def load_all_chunked_docs_from_mysql(
        cls, bot_id: str, version: str
    ) -> List["RetrieverDocument"]:
        faq_db_docs = mysql_pool.fetchall(
            "SELECT id, content, metadata, unix_timestamp(updated_at) as timestamp FROM qa_doc_faq WHERE qa_bot_id=%s and qa_bot_version=%s;",
            (bot_id, version),
        )
        faq_docs = []
        for doc in faq_db_docs:
            faq_docs.append(
                cls.faq_retreiver_doc(
                    id=doc["id"], text=doc["content"], metadata=doc["metadata"], bot_uid=get_bot_uid(bot_id, version), timestamp=doc["timestamp"]
                )
            )
        logger.info(f"Loaded {len(faq_docs)} faq docs")

        other_db_docs = mysql_pool.fetchall(
            "SELECT id, metadata, content, unix_timestamp(updated_at) as timestamp FROM qa_doc_other WHERE qa_bot_id=%s and qa_bot_version=%s;",
            (bot_id, version),
        )
        logger.info(f"Loaded {len(other_db_docs)} other docs")
        chunked_other_docs = cls.chunked_retriever_docs_from_db_docs(
            other_db_docs, RetrieverDocumentType.OTHER, get_bot_uid(bot_id, version)
        )
        logger.info(f"Chunked to {len(chunked_other_docs)} other docs")

        website_db_docs = mysql_pool.fetchall(
            "SELECT id, metadata, content, unix_timestamp(updated_at) as timestamp FROM qa_doc_website WHERE qa_bot_id=%s and qa_bot_version=%s and is_crawled=1 and is_error=0;",
            (bot_id, version),
        )
        logger.info(f"Loaded {len(website_db_docs)} website docs")
        chunked_website_docs = cls.chunked_retriever_docs_from_db_docs(
            website_db_docs, RetrieverDocumentType.WEBSITE, get_bot_uid(bot_id, version)
        )
        logger.info(f"Chunked to {len(chunked_website_docs)} website docs")

        return chunked_website_docs + chunked_other_docs + faq_docs

def embed_retriever_document(retriever_document: RetrieverDocument):
    return retriever_document.to_milvus_schema_dict_and_embed()

def get_bot_uid(bot_id: str, version: str):
    return f"{bot_id}__{version}"


# Copyright Sierra

import abc
import enum
from litellm import completion

from typing import Optional, List, Dict, Any, Union

class BaseUserSimulationEnv(abc.ABC):
    metadata = {}

    @abc.abstractmethod
    def reset(self, instruction: Optional[str] = None) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def step(self, content: str) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def get_total_cost(self) -> float:
        raise NotImplementedError


class HumanUserSimulationEnv(BaseUserSimulationEnv):
    def reset(self, instruction: str) -> str:
        return input(f"{instruction}\n")

    def step(self, content: str) -> str:
        return input(f"{content}\n")

    def get_total_cost(self) -> float:
        return 0


class LLMUserSimulationEnv(BaseUserSimulationEnv):
    def __init__(self, model: str, provider: str) -> None:
        super().__init__()
        self.messages: List[Dict[str, Any]] = []
        self.model = model
        self.provider = provider
        self.total_cost = 0.0
        self.reset()

    def generate_next_message(self, messages: List[Dict[str, Any]]) -> str:
        res = completion(
            model=self.model, custom_llm_provider=self.provider, messages=messages
        )
        message = res.choices[0].message
        self.messages.append(message.model_dump())
        self.total_cost = res._hidden_params["response_cost"]
        return message.content

    def build_system_prompt(self, instruction: Optional[str]) -> str:
        instruction_display = (
            ("\n\nInstruction: " + instruction + "\n")
            if instruction is not None
            else ""
        )
        return f"""You are a user interacting with an agent.{instruction_display}
Rules:
- Just generate one line at a time to simulate the user's message.
- Do not give away all the instruction at once. Only provide the information that is necessary for the current step.
- Do not hallucinate information that is not provided in the instruction. For example, if the agent asks for the order id but it is not mentioned in the instruction, do not make up an order id, just say you do not remember or have it.
- If the instruction goal is satisified, generate '###STOP###' as a standalone message without anything else to end the conversation.
- Do not repeat the exact instruction in the conversation. Instead, use your own words to convey the same information.
- Try to make the conversation as natural as possible, and stick to the personalities in the instruction."""

    def reset(self, instruction: Optional[str] = None) -> str:
        self.messages = [
            {
                "role": "system",
                "content": self.build_system_prompt(instruction=instruction),
            },
            {"role": "user", "content": "Hi! How can I help you today?"},
        ]
        return self.generate_next_message(self.messages)

    def step(self, content: str) -> str:
        self.messages.append({"role": "user", "content": content})
        return self.generate_next_message(self.messages)

    def get_total_cost(self) -> float:
        return self.total_cost


class ReactUserSimulationEnv(LLMUserSimulationEnv):
    def __init__(self, model: str, provider: str) -> None:
        super().__init__(model=model, provider=provider)
        self.reset()

    def build_system_prompt(self, instruction: Optional[str]) -> str:
        instruction_display = (
            ("\n\nInstruction: " + instruction + "\n")
            if instruction is not None
            else ""
        )
        return f"""You are a user interacting with an agent.{instruction_display}
Rules:
- First, generate a Thought about what to do next (this message will not be sent to the agent).
- Then, generate a one line User Response to simulate the user's message (this message will be sent to the agent).
- Do not give away all the instruction at once. Only provide the information that is necessary for the current step.
- Do not hallucinate information that is not provided in the instruction. For example, if the agent asks for the order id but it is not mentioned in the instruction, do not make up an order id, just say you do not remember or have it.
- If the instruction goal is satisified, generate '###STOP###' as the User Response without anything else to end the conversation.
- Do not repeat the exact instruction in the conversation. Instead, use your own words to convey the same information.
- Try to make the conversation as natural as possible, and stick to the personalities in the instruction.

Format:

Thought:
<the thought>

User Response:
<the user response (this will be parsed and sent to the agent)>"""

    def generate_next_message(self, messages: List[Dict[str, Any]]) -> str:
        res = completion(
            model=self.model, custom_llm_provider=self.provider, messages=messages
        )
        message = res.choices[0].message
        self.messages.append(message.model_dump())
        self.total_cost = res._hidden_params["response_cost"]
        return self.parse_response(message.content)

    def reset(self, instruction: Optional[str] = None) -> str:
        self.messages = [
            {
                "role": "system",
                "content": self.build_system_prompt(instruction=instruction),
            },
            {"role": "user", "content": "Hi! How can I help you today?"},
        ]
        return self.generate_next_message(self.messages)

    def parse_response(self, response: str) -> str:
        if "###STOP###" in response:
            return "###STOP###"
        elif "Thought:" in response:
            _, user_response = response.split("Thought:")
            return user_response.strip()
        elif "User Response:" in response:
            _, user_response = response.split("User Response:")
            return user_response.strip()
        else:
            raise ValueError(f"Invalid response format: {response}")

    def step(self, content: str) -> str:
        self.messages.append({"role": "user", "content": content})
        return self.generate_next_message(self.messages)

    def get_total_cost(self) -> float:
        return self.total_cost


class VerifyUserSimulationEnv(LLMUserSimulationEnv):
    def __init__(self, model: str, provider: str, max_attempts: int = 3) -> None:
        self.model = model
        self.provider = provider
        self.max_attempts = max_attempts
        self.reset()

    def generate_next_message(self, messages: List[Dict[str, Any]]) -> str:
        attempts = 0
        cur_message = None
        while attempts < self.max_attempts:
            res = completion(
                model=self.model, custom_llm_provider=self.provider, messages=messages
            )
            cur_message = res.choices[0].message
            self.total_cost = res._hidden_params["response_cost"]
            if verify(self.model, self.provider, cur_message, messages):
                self.messages.append(cur_message.model_dump())
                return cur_message.content
            attempts += 1
        assert cur_message is not None
        return cur_message.content

    def reset(self, instruction: Optional[str] = None) -> str:
        self.messages = [
            {
                "role": "system",
                "content": self.build_system_prompt(instruction=instruction),
            },
            {"role": "user", "content": "Hi! How can I help you today?"},
        ]
        return self.generate_next_message(self.messages)

    def step(self, content: str) -> str:
        self.messages.append({"role": "user", "content": content})
        return self.generate_next_message(self.messages)

    def get_total_cost(self) -> float:
        return self.total_cost


def map_role_label(role: str) -> str:
    if role == "user":
        return "Customer"
    elif role == "assistant":
        return "Agent"
    else:
        return role.capitalize()


def verify(
    model: str, provider: str, response: str, messages: List[Dict[str, Any]]
) -> bool:
    transcript = "\n".join(
        [
            f"{map_role_label(message['role'])}: {message['content']}"
            for message in messages
        ]
    )
    prompt = f"""You are a supervisor of the Agent in the conversation. You are given a Transcript of a conversation between a Customer and an Agent. The Customer has generated a Response, and you need to verify if it is satisfactory (true) or not (false).
Your answer will be parsed, so do not include any other text than the classification (true or false).
    
# Transcript:
{transcript}

# Response:
{response}

-----

Classification:"""
    res = completion(
        model=model,
        custom_llm_provider=provider,
        messages=[{"role": "user", "content": prompt}],
    )
    return "true" in res.choices[0].message.content.lower()


def reflect(
    model: str, provider: str, response: str, messages: List[Dict[str, Any]]
) -> str:
    transcript = "\n".join(
        [
            f"{map_role_label(message['role'])}: {message['content']}"
            for message in messages
        ]
    )
    prompt = f"""You are a supervisor of the Agent in the conversation. You are given a Transcript of a conversation between a (simulated) Customer and an Agent. The Customer generated a Response that was marked as unsatisfactory by you.
You need to generate a Reflection on what went wrong in the conversation, and propose a new Response that should fix the issues.
Your answer will be parsed, so do not include any other text than the classification (true or false).
    
# Transcript:
{transcript}

# Response:
{response}

# Format:

Reflection:
<the reflection>

Response:
<the response (this will be parsed and sent to the agent)>"""
    res = completion(
        model=model,
        custom_llm_provider=provider,
        messages=[{"role": "user", "content": prompt}],
    )
    _, response = res.choices[0].message.content.split("Response:")
    return response.strip()


class ReflectionUserSimulationEnv(LLMUserSimulationEnv):
    def __init__(self, model: str, provider: str, max_attempts: int = 2) -> None:
        self.model = model
        self.provider = provider
        self.max_attempts = max_attempts
        self.reset()

    def generate_next_message(self, messages: List[Dict[str, Any]]) -> str:
        cur_messages = messages.copy()
        initial_response = super().generate_next_message(cur_messages)
        if verify(self.model, self.provider, initial_response, cur_messages):
            return initial_response
        attempts = 1
        while attempts < self.max_attempts:
            new_message = reflect(
                self.model, self.provider, initial_response, cur_messages
            )
            cur_messages.append({"role": "user", "content": new_message})
            new_response = super().generate_next_message(cur_messages)
            if verify(self.model, self.provider, new_response, cur_messages):
                return new_response
            attempts += 1
        return initial_response

    def reset(self, instruction: Optional[str] = None) -> str:
        self.messages = [
            {
                "role": "system",
                "content": self.build_system_prompt(instruction=instruction),
            },
            {"role": "user", "content": "Hi! How can I help you today?"},
        ]
        return self.generate_next_message(self.messages)

    def step(self, content: str) -> str:
        self.messages.append({"role": "user", "content": content})
        return self.generate_next_message(self.messages)

    def get_total_cost(self) -> float:
        return self.total_cost


class UserStrategy(enum.Enum):
    HUMAN = "human"
    LLM = "llm"
    REACT = "react"
    VERIFY = "verify"
    REFLECTION = "reflection"


def load_user(
    user_strategy: Union[str, UserStrategy],
    model: Optional[str] = "gpt-4o",
    provider: Optional[str] = None,
) -> BaseUserSimulationEnv:
    if isinstance(user_strategy, str):
        user_strategy = UserStrategy(user_strategy)
    if user_strategy == UserStrategy.HUMAN:
        return HumanUserSimulationEnv()
    elif user_strategy == UserStrategy.LLM:
        if model is None:
            raise ValueError("LLM user strategy requires a model")
        if provider is None:
            raise ValueError("LLM user strategy requires a model provider")
        return LLMUserSimulationEnv(model=model, provider=provider)
    elif user_strategy == UserStrategy.REACT:
        if model is None:
            raise ValueError("React user strategy requires a model")
        if provider is None:
            raise ValueError("React user strategy requires a model provider")
        return ReactUserSimulationEnv(model=model, provider=provider)
    elif user_strategy == UserStrategy.VERIFY:
        if model is None:
            raise ValueError("Verify user strategy requires a model")
        if provider is None:
            raise ValueError("Verify user strategy requires a model provider")
        return VerifyUserSimulationEnv(model=model, provider=provider)
    elif user_strategy == UserStrategy.REFLECTION:
        if model is None:
            raise ValueError("Reflection user strategy requires a model")
        if provider is None:
            raise ValueError("Reflection user strategy requires a model provider")
        return ReflectionUserSimulationEnv(model=model, provider=provider)
    raise ValueError(f"Unknown user strategy {user_strategy}")



from typing import Any

from pydantic import BaseModel

from benchmark.tau_bench.model_utils.api.datapoint import Datapoint
from benchmark.tau_bench.model_utils.model.vllm_completion import VLLMCompletionModel
from benchmark.tau_bench.model_utils.model.vllm_utils import generate_request


class OutlinesCompletionModel(VLLMCompletionModel):
    def parse_force_from_prompt(
        self, prompt: str, typ: BaseModel, temperature: float | None = None
    ) -> dict[str, Any]:
        if temperature is None:
            temperature = self.temperature
        schema = typ.model_json_schema()
        res = generate_request(
            url=self.url,
            prompt=prompt,
            force_json=True,
            schema=schema,
            temperature=temperature,
        )
        return self.handle_parse_force_response(prompt=prompt, content=res)

    def get_approx_cost(self, dp: Datapoint) -> float:
        return super().get_approx_cost(dp)

    def get_latency(self, dp: Datapoint) -> float:
        return super().get_latency(dp)

    def get_capability(self) -> float:
        return super().get_capability()

    def supports_dp(self, dp: Datapoint) -> bool:
        return super().supports_dp(dp)



import os
import sys
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.append(root_dir)
from dotenv import load_dotenv

from benchmark.tau_bench.auto_error_identification import get_args, run_error_identification
load_dotenv()

if __name__ == "__main__":
    '''
    Provide --results-path and --output-dir
    '''
    sys.argv += ['--env', 'retail']
    sys.argv += ['--platform', 'openai']
    sys.argv += ['--max-concurrency', '16']
    args = get_args()
    run_error_identification(args)
    


from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs
from arklex.env.tools.shopify.utils_cart import *
from arklex.env.tools.shopify.utils_nav import *

from arklex.env.tools.tools import register_tool

description = "Add items to user's shopping cart."
slots = [
    ShopifySlots.CART_ID,
    ShopifySlots.ADD_LINE_ITEM
]
outputs = [
    ShopifyOutputs.CART_ADD_ITEMS_DETAILS
]
CART_ADD_ITEM_ERROR = "error: products could not be added to cart"
errors = [CART_ADD_ITEM_ERROR]

@register_tool(description, slots, outputs, lambda x: x not in errors)
def cart_add_items(cart_id, add_line_items: list, **kwargs):
    auth = authorify_storefront(kwargs)
    if auth["error"]:
        return auth["error"]
    
    variable = {
        "cartId": cart_id,
        "lines": [
            {"merchandiseId": item,
                "quantity": 1
            } for item in add_line_items
        ]
    }
    headers = {
        "X-Shopify-Storefront-Access-Token": auth["storefront_token"]
    }
    query = '''
    mutation cartLinesAdd($cartId: ID!, $lines: [CartLineInput!]!) {
        cartLinesAdd(cartId: $cartId, lines: $lines) {
            cart {
                checkoutUrl
            }
        }
    }
    '''
    response = requests.post(auth["storefront_url"], json={"query": query, "variables": variable}, headers=headers)
    if response.status_code == 200:
        cart_data = response.json()
        if "errors" in cart_data:
            return CART_ADD_ITEM_ERROR
        else:
            return "Items are successfully added to the shopping cart. " + json.dumps(cart_data["data"]["cartLinesAdd"]["cart"])
    else:
        return CART_ADD_ITEM_ERROR



import logging
import os
from typing import Any, Iterator, Union

from langgraph.graph import StateGraph, START
from langchain_openai import ChatOpenAI

from arklex.env.workers.worker import BaseWorker, register_worker
from arklex.utils.graph_state import MessageState
from arklex.env.tools.utils import ToolGenerator
from arklex.env.tools.RAG.retrievers.faiss_retriever import RetrieveEngine
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import PROVIDER_MAP


logger = logging.getLogger(__name__)


@register_worker
class FaissRAGWorker(BaseWorker):

    description = "Answer the user's questions based on the company's internal documentations (unstructured text data), such as the policies, FAQs, and product information"

    def __init__(self,
                 # stream_ reponse is a boolean value that determines whether the response should be streamed or not.
                 # i.e in the case of RagMessageWorker it should be set to false.
                 stream_response: bool = True):
        super().__init__()
        self.action_graph = self._create_action_graph()
        self.llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000
        )
        self.stream_response = stream_response

    def choose_tool_generator(self, state: MessageState):
        if self.stream_response and state["is_stream"]:
            return "stream_tool_generator"
        return "tool_generator"

    def _create_action_graph(self):
        workflow = StateGraph(MessageState)
        # Add nodes for each worker
        workflow.add_node("retriever", RetrieveEngine.faiss_retrieve)
        workflow.add_node("tool_generator", ToolGenerator.context_generate)
        workflow.add_node("stream_tool_generator", ToolGenerator.stream_context_generate)

        # Add edges
        workflow.add_edge(START, "retriever")
        workflow.add_conditional_edges(
            "retriever", self.choose_tool_generator)
        return workflow

    def execute(self, msg_state: MessageState):
        graph = self.action_graph.compile()
        result = graph.invoke(msg_state)
        return result



import abc
from typing import Any, TypeVar

from pydantic import BaseModel

from benchmark.tau_bench.model_utils.api.datapoint import (
    BinaryClassifyDatapoint,
    ClassifyDatapoint,
    GenerateDatapoint,
    ParseDatapoint,
    ParseForceDatapoint,
    ScoreDatapoint,
)
from benchmark.tau_bench.model_utils.api.types import PartialObj
from benchmark.tau_bench.model_utils.model.model import (
    BinaryClassifyModel,
    ClassifyModel,
    GenerateModel,
    ParseForceModel,
    ParseModel,
    Platform,
    ScoreModel,
)

T = TypeVar("T", bound=BaseModel)

LLM_SAMPLING_TEMPERATURE_EPS = 1e-5


def wrap_temperature(temperature: float) -> float:
    return max(temperature, LLM_SAMPLING_TEMPERATURE_EPS)


class GeneralModel(
    ClassifyModel,
    BinaryClassifyModel,
    ParseModel,
    GenerateModel,
    ParseForceModel,
    ScoreModel,
):
    @abc.abstractmethod
    def classify(
        self,
        instruction: str,
        text: str,
        options: list[str],
        examples: list[ClassifyDatapoint] | None = None,
        temperature: float | None = None,
    ) -> int:
        raise NotImplementedError

    def binary_classify(
        self,
        instruction: str,
        text: str,
        examples: list[BinaryClassifyDatapoint] | None = None,
        temperature: float | None = None,
    ) -> bool:
        return (
            self.classify(
                instruction,
                text,
                ["true", "false"],
                examples=(
                    None
                    if examples is None
                    else [
                        ClassifyDatapoint(
                            instruction=example.instruction,
                            text=example.text,
                            options=["true", "false"],
                            response=0 if example.response else 1,
                        )
                        for example in examples
                    ]
                ),
                temperature=temperature,
            )
            == 0
        )

    @abc.abstractmethod
    def parse(
        self,
        text: str,
        typ: type[T] | dict[str, Any],
        examples: list[ParseDatapoint] | None = None,
        temperature: float | None = None,
    ) -> T | PartialObj | dict[str, Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def generate(
        self,
        instruction: str,
        text: str,
        examples: list[GenerateDatapoint] | None = None,
        temperature: float | None = None,
    ) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def parse_force(
        self,
        instruction: str,
        typ: type[T] | dict[str, Any],
        text: str | None = None,
        examples: list[ParseForceDatapoint] | None = None,
        temperature: float | None = None,
    ) -> T | dict[str, Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def score(
        self,
        instruction: str,
        text: str,
        min: int,
        max: int,
        examples: list[ScoreDatapoint] | None = None,
        temperature: float | None = None,
    ) -> int:
        raise NotImplementedError


def default_model() -> GeneralModel:
    from benchmark.tau_bench.model_utils.model.openai import OpenAIModel

    return OpenAIModel()


def default_quick_model() -> GeneralModel:
    from benchmark.tau_bench.model_utils.model.openai import OpenAIModel

    return OpenAIModel(model="gpt-4o-mini")


def model_factory(
    model_id: str,
    platform: str | Platform,
    base_url: str | None = None,
    api_key: str | None = None,
    temperature: float = 0.0,
) -> GeneralModel:
    if isinstance(platform, str):
        platform = Platform(platform)
    if platform == Platform.OPENAI:
        from benchmark.tau_bench.model_utils.model.openai import OpenAIModel

        return OpenAIModel(model=model_id, api_key=api_key, temperature=temperature)
    elif platform == Platform.MISTRAL:
        from benchmark.tau_bench.model_utils.model.mistral import MistralModel

        return MistralModel(model=model_id, api_key=api_key, temperature=temperature)
    elif platform == Platform.ANTHROPIC:
        from benchmark.tau_bench.model_utils.model.claude import ClaudeModel

        return ClaudeModel(model=model_id, api_key=api_key, temperature=temperature)

    elif platform == Platform.ANYSCALE:
        from benchmark.tau_bench.model_utils.model.anyscale import AnyscaleModel

        return AnyscaleModel(model=model_id, api_key=api_key, temperature=temperature)
    elif platform == Platform.OUTLINES:
        if base_url is None:
            raise ValueError("base_url must be provided for custom models")
        from benchmark.tau_bench.model_utils.model.outlines_completion import OutlinesCompletionModel

        return OutlinesCompletionModel(model=model_id, base_url=base_url, temperature=temperature)
    elif platform == Platform.VLLM_CHAT:
        if base_url is None:
            raise ValueError("base_url must be provided for custom models")
        from benchmark.tau_bench.model_utils.model.vllm_chat import VLLMChatModel

        return VLLMChatModel(
            model=model_id,
            base_url=base_url,
            api_key="sk-no-api-key-required" if api_key is None else api_key,
            temperature=temperature,
        )
    else:
        if base_url is None:
            raise ValueError("base_url must be provided for custom models")
        from benchmark.tau_bench.model_utils.model.vllm_completion import VLLMCompletionModel

        return VLLMCompletionModel(model=model_id, base_url=base_url, temperature=temperature)



import json
from typing import Any, Dict

import shopify

# general GraphQL navigation utilities
from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs
from arklex.env.tools.shopify.utils import authorify_admin
from arklex.env.tools.shopify.utils_nav import *

from arklex.env.tools.tools import register_tool

description = "Get the details of the collection."
slots = [
    ShopifySlots.COLLECTION_ID,
    *PAGEINFO_SLOTS
]
outputs = [
    ShopifyOutputs.COLLECTIONS_DETAILS,
    *PAGEINFO_OUTPUTS
]
COLLECTION_NOT_FOUND = "error: collection not found"
errors = [COLLECTION_NOT_FOUND]

@register_tool(description, slots, outputs, lambda x: x not in errors)
def get_collection(collection_id: list, **kwargs) -> str:
    nav = cursorify(kwargs)
    if not nav[1]:
        return nav[0]
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]
    
    try:
        with shopify.Session.temp(**auth["value"]):
            response = shopify.GraphQL().execute(f"""
            {{
                collection (id: "{collection_id}") {{
                    title
                    description
                    productsCount {{
                        count
                    }}
                    products ({nav[0]}) {{
                        nodes {{
                            id
                        }}
                        pageInfo {{
                            endCursor
                            hasNextPage
                            hasPreviousPage
                            startCursor
                        }}
                    }}
                }}
            }}
            """)
            results = json.loads(response)["data"]["collection"]
            return results
    except Exception as e:
        return {"content": COLLECTION_NOT_FOUND}


import logging

from langgraph.graph import StateGraph, START
from langchain_openai import ChatOpenAI
from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser

from arklex.env.workers.worker import BaseWorker, register_worker
from arklex.env.prompts import load_prompts
from arklex.env.tools.utils import ToolGenerator
from arklex.env.tools.database.utils import DatabaseActions
from arklex.utils.utils import chunk_string
from arklex.utils.graph_state import MessageState
from arklex.utils.model_config import MODEL



logger = logging.getLogger(__name__)


@register_worker
class DataBaseWorker(BaseWorker):

    description = "Help the user with actions related to customer support like a booking system with structured data, always involving search, insert, update, and delete operations."

    def __init__(self):
        self.llm = ChatOpenAI(model=MODEL["model_type_or_path"], timeout=30000)
        self.actions = {
            "SearchShow": "Search for shows", 
            "BookShow": "Book a show", 
            "CheckBooking": "Check details of booked show(s)",
            "CancelBooking": "Cancel a booking",
            "Others": "Other actions not mentioned above"
        }
        self.DBActions = DatabaseActions()
        self.action_graph = self._create_action_graph()

    def search_show(self, state: MessageState):
        return self.DBActions.search_show(state)
    
    def book_show(self, state: MessageState):
        return self.DBActions.book_show(state)
    
    def check_booking(self, state: MessageState):
        return self.DBActions.check_booking(state)
    
    def cancel_booking(self, state: MessageState):
        return self.DBActions.cancel_booking(state)

    def verify_action(self, msg_state: MessageState):
        user_intent = msg_state["orchestrator_message"].attribute.get("task", "")
        actions_info = "\n".join([f"{name}: {description}" for name, description in self.actions.items()])
        actions_name = ", ".join(self.actions.keys())

        prompts = load_prompts(msg_state["bot_config"])
        prompt = PromptTemplate.from_template(prompts["database_action_prompt"])
        input_prompt = prompt.invoke({"user_intent": user_intent, "actions_info": actions_info, "actions_name": actions_name})
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        logger.info(f"Chunked prompt for deciding choosing DB action: {chunked_prompt}")
        final_chain = self.llm | StrOutputParser()
        try:
            answer = final_chain.invoke(chunked_prompt)
            for action_name in self.actions.keys():
                if action_name in answer:
                    logger.info(f"Chosen action in the database worker: {action_name}")
                    return action_name
            logger.info(f"Base action chosen in the database worker: Others")
            return "Others"
        except Exception as e:
            logger.error(f"Error occurred while choosing action in the database worker: {e}")
            return "Others"

        
    def _create_action_graph(self):
        workflow = StateGraph(MessageState)
        # Add nodes for each worker
        workflow.add_node("SearchShow", self.search_show)
        workflow.add_node("BookShow", self.book_show)
        workflow.add_node("CheckBooking", self.check_booking)
        workflow.add_node("CancelBooking", self.cancel_booking)
        workflow.add_node("Others", ToolGenerator.generate)
        workflow.add_node("tool_generator", ToolGenerator.context_generate)
        workflow.add_conditional_edges(START, self.verify_action)
        workflow.add_edge("SearchShow", "tool_generator")
        workflow.add_edge("BookShow", "tool_generator")
        workflow.add_edge("CheckBooking", "tool_generator")
        workflow.add_edge("CancelBooking", "tool_generator")
        return workflow

    def execute(self, msg_state: MessageState):
        self.DBActions.log_in()
        msg_state["slots"] = self.DBActions.init_slots(msg_state["slots"], msg_state["bot_config"])
        graph = self.action_graph.compile()
        result = graph.invoke(msg_state)
        return result



from langchain_openai import ChatOpenAI
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_anthropic import ChatAnthropic
from langchain_openai import OpenAIEmbeddings
from langchain_google_genai import GoogleGenerativeAIEmbeddings
from langchain_huggingface.embeddings import HuggingFaceEmbeddings

LLM_PROVIDERS = ["openai", "gemini", "anthropic"]

PROVIDER_MAP = {
    "anthropic": ChatAnthropic,
    "gemini": ChatGoogleGenerativeAI,
    "openai": ChatOpenAI  
}

PROVIDER_EMBEDDINGS = {
    "anthropic": HuggingFaceEmbeddings,
    "gemini": GoogleGenerativeAIEmbeddings,
    "openai": OpenAIEmbeddings 
}
PROVIDER_EMBEDDING_MODELS = {
    "anthropic": "sentence-transformers/all-mpnet-base-v2",
    "gemini": "models/embedding-001",
    "openai": "text-embedding-ada-002"
}


generate_tasks_sys_prompt = """The builder plans to create a chatbot designed to fulfill user's objectives. Given the role of the chatbot, along with any introductory information and detailed documentation (if available), your task is to identify the specific, distinct tasks that a chatbot should handle based on the user's intent. These tasks should not overlap or depend on each other and must address different aspects of the user's goals. Ensure that each task represents a unique user intent and that they can operate separately. Return the response in JSON format.

For Example:

Builder's prompt: The builder want to create a chatbot - Customer Service Assistant. The customer service assistant typically handles tasks such as answering customer inquiries, making product recommendations, assisting with orders, processing returns and exchanges, supporting billing and payments, addressing complaints, and managing customer accounts.
Builder's Information: Amazon.com is a large e-commerce platform that sells a wide variety of products, ranging from electronics to groceries.
Builder's documentations: 
https://www.amazon.com/
Holiday Deals
Disability Customer Support
Same-Day Delivery
Medical Care
Customer Service
Amazon Basics
Groceries
Prime
Buy Again
New Releases
Pharmacy
Shop By Interest
Amazon Home
Amazon Business
Subscribe & Save
Livestreams
luwanamazon's Amazon.com
Best Sellers
Household, Health & Baby Care
Sell
Gift Cards

https://www.amazon.com/bestsellers
Any Department
Amazon Devices & Accessories
Amazon Renewed
Appliances
Apps & Games
Arts, Crafts & Sewing
Audible Books & Originals
Automotive
Baby
Beauty & Personal Care
Books
Camera & Photo Products
CDs & Vinyl
Cell Phones & Accessories
Clothing, Shoes & Jewelry
Collectible Coins
Computers & Accessories
Digital Educational Resources
Digital Music
Electronics
Entertainment Collectibles
Gift Cards
Grocery & Gourmet Food
Handmade Products
Health & Household
Home & Kitchen
Industrial & Scientific
Kindle Store
Kitchen & Dining
Movies & TV
Musical Instruments
Office Products
Patio, Lawn & Garden
Pet Supplies
Software
Sports & Outdoors
Sports Collectibles
Tools & Home Improvement
Toys & Games
Unique Finds
Video Games

Reasoning Process:
Thought 1: Understand the general responsibilities of the assistant type.
Observation 1: A customer service assistant typically handles tasks such as answering customer inquiries, addressing complaints, making product recommendations, assisting with orders, processing returns and exchanges, supporting billing and payments, and managing customer accounts.

Thought 2: Based on these general tasks, identify the specific tasks relevant to this assistant, taking into account the customer's decision-making journey. Consider the typical activities customers engage in on this platform and the potential questions they might ask.
Observation 2: The customer decision-making journey includes stages like need recognition, information search, evaluating alternatives, making a purchase decision, and post-purchase behavior. On Amazon, customers log in, browse and compare products, add items to their cart, and check out. They also track orders, manage returns, and leave reviews. Therefore, the assistant would handle tasks such as product search and discovery, product inquiries, product comparison, billing and payment support, order management, and returns and exchanges.

Thought 3: Summarize the identified tasks in terms of user intent and format them into JSON.
Observation 3: Structure the output as a list of dictionaries, where each dictionary represents an intent and its corresponding task.

Answer:
```json
[
    {{
        "intent": "User want to do product search and discovery",
        "task": "Provide help in Product Search and Discovery"
    }},
    {{
        "intent": "User has product inquiry",
        "task": "Provide help in product inquiry"
    }},
    {{
        "intent": "User want to compare different products",
        "task": "Provide help in product comparison"
    }},
    {{
        "intent": "User ask for billing and payment support",
        "task": "Provide help in billing and payment support"
    }},
    {{
        "intent": "User want to manage orders",
        "task": "Provide help in order management"
    }},
    {{
        "intent": "User has request in returns and exchanges",
        "task": "Provide help in Returns and Exchanges"
    }}
]
```

Builder's prompt: The builder want to create a chatbot - {role}. {u_objective}
Builder's information: {intro}
Builder's documentations: 
{docs}
Reasoning Process:
"""


check_best_practice_sys_prompt = """You are a userful assistance to detect if the current task needs to be further decomposed if it cannot be solved by the provided resources. Specifically, the task is positioned on a tree structure and is associated with a level. Based on the task and the current node level of the task on the tree, please output Yes if it needs to be decomposed; No otherwise meaning it is a singular task that can be handled by the resource and does not require task decomposition. Please also provide explanations for your choice. 

Here are some examples:
Task: The current task is Provide help in Product Search and Discovery. The current node level of the task is 1. 
Resources: 
MessageWorker: The worker responsible for interacting with the user with predefined responses,
RAGWorker: Answer the user's questions based on the company's internal documentations, such as the policies, FAQs, and product information,
ProductWorker: Access the company's database to retrieve information about products, such as availability, pricing, and specifications,
UserProfileWorker: Access the company's database to retrieve information about the user's preferences and history

Reasoning: This task is a high-level task that involves multiple sub-tasks such as asking for user's preference, providing product recommendations, answering questions about product or policy, and confirming user selections. Each sub-task requires different worker to complete. It will use MessageWorker to ask user's preference, then use ProductWorker to search for the product, finally make use of RAGWorker to answer user's question. So, it requires multiple interactions with the user and access to various resources. Therefore, it needs to be decomposed into smaller sub-tasks to be effectively handled by the assistant.
Answer: 
```json
{{
    "answer": "Yes"
}}
```

Task: The current task is booking a broadway show ticket. The current node level of the task is 1.
Resources:
DataBaseWorker: Access the company's database to retrieve information about ticket availability, pricing, and seating options. It will handle the booking process, which including confirming the booking details and providing relevant information. It can also handle the cancel process.
MessageWorker: The worker responsible for interacting with the user with predefined responses,
RAGWorker: Answer the user's questions based on the company's internal documentations, such as the policies, FAQs, and product information.

Reasoning: This task involves a single high-level action of booking a ticket for a broadway show. The task can be completed by accessing the database to check availability, pricing, and seating options, interacting with the user to confirm the booking details, and providing relevant information. Since it is a singular task that can be handled by the single resource without further decomposition, the answer is No.
Answer: 
```json
{{
    "answer": "No"
}}
```

Task: The current task is {task}. The current node level of the task is {level}.
Resources: {resources}
Reasoning:
"""


generate_best_practice_sys_prompt = """Given the background information about the chatbot, the task it needs to handle, and the available resources, your task is to generate a step-by-step best practice for addressing this task. Each step should represent a distinct interaction with the user, where the next step builds upon the user's response. Avoid breaking down sequences of internal worker actions within a single turn into multiple steps. Return the answer in JSON format. Only use resources listed in the input, resources listed in the Example may not be available.

For example:
Background: The builder want to create a chatbot - Customer Service Assistant. The customer service assistant typically handles tasks such as answering customer inquiries, making product recommendations, assisting with orders, processing returns and exchanges, supporting billing and payments, addressing complaints, and managing customer accounts.

Task: Provide help in Product Search and Discovery

Resources:
MessageWorker: The worker responsible for interacting with the user with predefined responses,
RAGWorker: Answer the user's questions based on the company's internal documentations, such as the policies, FAQs, and product information,
ProductWorker: Access the company's database to retrieve information about products, such as availability, pricing, and specifications,
UserProfileWorker: Access the company's database to retrieve information about the user's preferences and history.

Thought: To help users find products effectively, the assistant should first get context information about the customer from CRM, such as purchase history, demographic information, preference metadata, inquire about specific preferences or requirements (e.g., brand, features, price range) specific for the request. Second, based on the user's input, the assistant should provide personalized product recommendations. Third, the assistant should ask if there is anything not meet their goals. Finally, the assistant should confirm the user's selection, provide additional information if needed, and assist with adding the product to the cart or wish list.
Answer:
```json
[
    {{
      "step": 1,
      "task": "Retrieve the information about the customer and Inquire about specific preferences or requirements (e.g., brand, features, price range)."
    }},
    {{
      "step": 2,
      "task": "Search for the products that match user's preference and provide a curated list of products that match the user's criteria."
    }},
    {{
      "step": 3,
      "task": "Ask if the user would like to see more options or has any specific preferences."
    }},
    {{
      "step": 4,
      "task": "Confirm if the user is ready to proceed with a purchase or needs more help."
    }},
    {{
      "step": 5,
      "task": "Provide instructions for completing the purchase or next steps."
    }}
]
```

Background: The builder want to create a chatbot - {role}. {u_objective}
Task: {task}
Resources: {resources}
Thought:
"""


# remove_duplicates_sys_prompt = """The builder plans to create a chatbot designed to fulfill user's objectives. Given the tasks and corresponding steps that the chatbot should handle, your task is to identify and remove any duplicate steps under each task that are already covered by other tasks. Ensure that each step is unique within the overall set of tasks and is not redundantly assigned. Return the response in JSON format.

# Tasks: {tasks}
# Answer:
# """

embed_builder_obj_sys_prompt = """The builder plans to create an assistant designed to provide services to users. Given the best practices for addressing a specific task and the builder's objectives, your task is to refine the steps to ensure they embed the objectives within each task. Return the answer in JSON format.

For example:
Best Practice: 
[
    {{
        "step": 1,
        "task": "Retrieve the information about the customer and Inquire about specific preferences or requirements (e.g., brand, features, price range)."
    }},
    {{
        "step": 2,
        "task": "Provide a curated list of products that match the user's criteria."
    }},
    {{
        "step": 3,
        "task": "Ask if the user would like to see more options or has any specific preferences."
    }},
    {{
        "step": 4,
        "task": "Confirm if the user is ready to proceed with a purchase or needs more help."
    }},
    {{
        "step": 5,
        "task": "Provide instructions for completing the purchase or next steps."
    }}
]
Build's objective: The customer service assistant helps in persuading customer to sign up the Prime membership.
Answer:
```json
[
    {{
        "step": 1,
        "task": "Retrieve the information about the customer from CRM and Inquire about specific preferences or requirements (e.g., brand, features, price range)."
    }},
    {{
        "step": 2,
        "task": "Provide a curated list of products that match the user's criteria."
    }},
    {{
        "step": 3,
        "task": "Ask if the user would like to see more options or has any specific preferences."
    }},
    {{
        "step": 4,
        "task": "Confirm if the user is ready to proceed with a purchase or needs more help."
    }},
    {{
        "step": 5,
        "task": "Persuade the user to sign up for the Prime membership."
    }}
]
```

Best Practice: {best_practice}
Build's objective: {b_objective}
Answer:
"""


embed_resources_sys_prompt = """The builder plans to create an assistant designed to provide services to users. Given the best practices for addressing a specific task, and the available resources, your task is to map the steps with the resources. The response should include the resources used for each step and example responses, if applicable. Return the answer in JSON format. Do not add any comment on the answer.

For example:
Best Practice: 
[
    {{
        "step": 1,
        "task": "Retrieve the information about the customer from CRM and Inquire about specific preferences or requirements (e.g., brand, features, price range)."
    }},
    {{
        "step": 2,
        "task": "Provide a curated list of products that match the user's criteria."
    }},
    {{
        "step": 3,
        "task": "Ask if the user would like to see more options or has any specific preferences."
    }},
    {{
        "step": 4,
        "task": "Confirm if the user is ready to proceed with a purchase or needs more help."
    }},
    {{
        "step": 5,
        "task": "Persuade the user to sign up for the Prime membership."
    }}
]
Resources:
{{
    "MessageWorker": "The worker responsible for interacting with the user with predefined responses",
    "RAGWorker": "Answer the user's questions based on the company's internal documentations, such as the policies, FAQs, and product information",
    "ProductWorker": "Access the company's database to retrieve information about products, such as availability, pricing, and specifications",
    "UserProfileWorker": "Access the company's database to retrieve information about the user's preferences and history"
}}
Answer:
```json
[
    {{
        "step": 1,
        "task": "Retrieve the information about the customer from CRM and Inquire about specific preferences or requirements (e.g., brand, features, price range)."
        "resource": "UserProfileWorker",
        "example_response": "Do you have some specific preferences or requirements for the product you are looking for?"
    }},
    {{
        "step": 2,
        "task": "Provide a curated list of products that match the user's criteria."
        "resource": "ProductWorker",
        "example_response": ""
    }},
    {{
        "step": 3,
        "task": "Ask if the user would like to see more options or has any specific preferences."
        "resource": "MessageWorker",
        "example_response": "Would you like to see more options or do you have any specific preferences?"
    }},
    {{
        "step": 4,
        "task": "Confirm if the user is ready to proceed with a purchase or needs more help."
        "resource": "MessageWorker",
        "example_response": "Are you ready to proceed with the purchase or do you need more help?"
    }},
    {{
        "step": 5,
        "task": "Persuade the user to sign up for the Prime membership."
        "resource": "MessageWorker",
        "example_response": "I noticed that you are a frequent shopper. Have you considered signing up for our Prime membership to enjoy exclusive benefits and discounts?"
    }}
]
```

Best Practice: {best_practice}
Resources: {resources}
Answer:
"""


generate_start_msg = """The builder plans to create a chatbot designed to fulfill user's objectives. Given the role of the chatbot, your task is to generate a starting message for the chatbot. Return the response in JSON format.

For Example:

Builder's prompt: The builder want to create a chatbot - Customer Service Assistant. The customer service assistant typically handles tasks such as answering customer inquiries, making product recommendations, assisting with orders, processing returns and exchanges, supporting billing and payments, addressing complaints, and managing customer accounts.
Start Message:
```json
{{
    "message": "Welcome to our Customer Service Assistant! How can I help you today?"
}}
```

Builder's prompt: The builder want to create a chatbot - {role}. {u_objective}
Start Message:
"""


# Copyright Sierra

import abc
from typing import Optional
from benchmark.tau_bench.envs.base import Env
from benchmark.tau_bench.types import SolveResult


class Agent(abc.ABC):
    @abc.abstractmethod
    def solve(
        self, env: Env, task_index: Optional[int] = None, max_num_steps: int = 30
    ) -> SolveResult:
        raise NotImplementedError



from ..tools import register_tool
from .utils import *

import datetime
import uuid
import pandas as pd

@register_tool(
    "Books an existing show",
    [{**SLOTS['show_name'], 'required': False}, 
     {**SLOTS['date'], 'required': False},
     {**SLOTS['time'], 'required': False},
     {**SLOTS['location'], 'required': False},
    ],
    [{
        "name": "query_result",
        "type": "str",
        "description": "A list of available shows that satisfies the given criteria (displays the first 10 results). If no show satisfies the criteria, returns 'No shows exist'",
    }],
    lambda x: x and x not in (LOG_IN_FAILURE, NO_SHOW_MESSAGE, MULTIPLE_SHOWS_MESSAGE)
)
def book_show(show_name=None, date=None, time=None, location=None) -> str | None:
    if not log_in(): return LOG_IN_FAILURE
    
    logger.info("Enter book show function")
    conn = sqlite3.connect(booking.db_path)
    cursor = conn.cursor()
    query = "SELECT id, show_name, date, time, description, location, price FROM show WHERE 1 = 1"
    params = []
    slots = {"show_name": show_name, "date": date, "time": time, "location": location}
    logger.info(f"{slots=}")
    for slot_name, slot_value in slots.items():
        if slot_value:
            query += f" AND {slot_name} = ?"
            params.append(slot_value)
            
    # Execute the query
    cursor.execute(query, params)
    rows = cursor.fetchall()
    logger.info(f"Rows found: {len(rows)}")
    
    response = None
    # Check whether info is enough to book a show
    if len(rows) == 0:
        response = NO_SHOW_MESSAGE
    elif len(rows) > 1:
        response = MULTIPLE_SHOWS_MESSAGE
    else:
        column_names = [column[0] for column in cursor.description]
        results = dict(zip(column_names, rows[0]))
        show_id = results["id"]

        # Insert a row into the booking table
        cursor.execute('''
            INSERT INTO booking (id, show_id, user_id, created_at)
            VALUES (?, ?, ?, ?)
        ''', ("booking_" + str(uuid.uuid4()),  show_id, booking.user_id, datetime.now()))

        results_df = pd.DataFrame([results])
        response = "The booked show is:\n" + results_df.to_string(index=False)
        
    cursor.close()
    conn.close()
    return response



# Copyright Sierra
import os
import json
from copy import deepcopy
from typing import List, Optional, Dict, Any

from arklex.orchestrator.orchestrator import AgentOrg
from arklex.env.env import Env

# from benchmark.tau_bench.envs.base import Env
from benchmark.tau_bench.agents.base import Agent
from benchmark.tau_bench.types import SolveResult, Action, RESPOND_ACTION_NAME


class AgentFirstOrg(Agent):
    def __init__(self, taskgraph_dir: str):
        self.taskgraph_dir = taskgraph_dir
        self.taskgraph_path = os.path.join(self.taskgraph_dir, "taskgraph.json")
        from tau_bench_eval import TauBenchResourceInitializer
        with open(self.taskgraph_path) as taskgraph:
            taskgraph = json.load(taskgraph)
            tau_bench_resource_initializer = TauBenchResourceInitializer()
            self.env = Env(
                tools = taskgraph.get("tools", []),
                workers = taskgraph.get("workers", []),
                slotsfillapi = taskgraph["slotfillapi"],
                resource_inizializer = tau_bench_resource_initializer
            )

            self.start_message = None
            for node in taskgraph['nodes']:
                if node[1].get("type", "") == 'start':
                    self.start_message = node[1]['attribute']["value"]
                    break

    def get_api_bot_response(self, history, user_text, parameters):
        data = {"text": user_text, 'chat_history': history, 'parameters': parameters}
        orchestrator = AgentOrg(config=self.taskgraph_path, env=self.env)
        result = orchestrator.get_response(data)
        return result['answer'], result['parameters']

    def solve(
        self, env, task_index: Optional[int] = None, max_num_steps: int = 30
    ) -> SolveResult:
        total_cost = 0.0
        env_reset_res = env.reset(task_index=task_index)
        obs = env_reset_res.observation
        info = env_reset_res.info.model_dump()
        reward = 0.0
        history = [
            {"role": "assistant", "content": self.start_message}
        ]
        messages = [
            {"role": "assistant", "content": self.start_message}
        ]
        params = {}
        user_text = obs
        message_index = 1
        
        for _ in range(max_num_steps):
            new_messages = []
            output, params = self.get_api_bot_response(deepcopy(history), user_text, params)

            user_message = {"role": "user", "content": user_text}
            assistant_message = {"role": "assistant", "content": output}
            history.append(user_message)
            history.append(assistant_message)

            while message_index < len(params["history"]):
                msg = params["history"][message_index]
                
                if not is_message_worker(msg):

                    if is_assistant_with_tool_calls(msg) or \
                        is_user(msg) or \
                        is_tool(msg):
                        new_messages.append(msg)
                        
                    if is_assistant_with_tool_calls(msg):
                        action = message_to_action(msg)
                        env_response = env.step(action)
                        reward = env_response.reward
                        info = {**info, **env_response.info.model_dump()}

                message_index += 1
            
            # total_cost += res._hidden_params["response_cost"]
            new_messages.append(assistant_message)
            action = message_to_action(assistant_message)
            env_response = env.step(action)
            reward = env_response.reward
            info = {**info, **env_response.info.model_dump()}
            
            user_text = env_response.observation

            if env_response.done:
                user_message = {"role": "user", "content": user_text}
                new_messages.append(user_message)
            messages.extend(new_messages)
            if env_response.done:
                break
        return SolveResult(
            reward=reward,
            info=info,
            messages=messages,
            total_cost=total_cost,
        )


def is_user(message):
    if message.get("role") == "user":
        return True
    return False

def is_tool(message):
    if message.get("role") == "tool":
        return True
    return False

def is_assistant_with_tool_calls(message):
    if message.get("role") != "assistant": return False
    if "tool_calls" not in message: return False
    if message["tool_calls"] is None: return False
    if len(message["tool_calls"]) == 0: return False
    if "function" not in message["tool_calls"][0]: return False
    if message["tool_calls"][0]["function"] is None: return False
    return True

def is_message_worker(message):
    if message.get("name") == "MessageWorker": return True
    if "tool_calls" not in message: return False
    if message["tool_calls"] is None: return False
    if len(message["tool_calls"]) == 0: return False
    if "function" not in message["tool_calls"][0]: return False
    if message["tool_calls"][0]["function"] is None: return False
    return message["tool_calls"][0]["function"].get("name") == "MessageWorker"



def message_to_action(
    message: Dict[str, Any],
) -> Action:
    if "tool_calls" in message and \
            message["tool_calls"] is not None and \
            len(message["tool_calls"]) > 0 and \
            message["tool_calls"][0]["function"] is not None:
        
        tool_call = message["tool_calls"][0]
        return Action(
            name=tool_call["function"]["name"],
            kwargs=json.loads(tool_call["function"]["arguments"]),
        )
    else:
        return Action(name=RESPOND_ACTION_NAME, kwargs={"content": message["content"]})




import logging

from langgraph.graph import StateGraph, START
from langchain_openai import ChatOpenAI


from arklex.env.workers.worker import BaseWorker, register_worker
from arklex.utils.graph_state import MessageState
from arklex.env.tools.utils import ToolGenerator
from arklex.env.tools.RAG.search import SearchEngine
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import PROVIDER_MAP


logger = logging.getLogger(__name__)


@register_worker
class SearchWorker(BaseWorker):

    description = "Answer the user's questions based on real-time online search results"

    def __init__(self):
        super().__init__()
        self.action_graph = self._create_action_graph()
        self.llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000
        )
     
    def _create_action_graph(self):
        workflow = StateGraph(MessageState)
        # Add nodes for each worker
        search_engine = SearchEngine()
        workflow.add_node("search_engine", search_engine.search)
        workflow.add_node("tool_generator", ToolGenerator.context_generate)
        # Add edges
        workflow.add_edge(START, "search_engine")
        workflow.add_edge("search_engine", "tool_generator")
        return workflow

    def execute(self, msg_state: MessageState):
        graph = self.action_graph.compile()
        result = graph.invoke(msg_state)
        return result



from typing import TypedDict, Annotated, Any, Optional, Union, List
import janus
from pydantic import BaseModel
from enum import Enum


### Bot-related classes
class BotConfig(BaseModel):
    bot_id: str
    version: str
    language: str
    bot_type: str
    available_workers: list[dict[str, Optional[str]]]


### Message-related classes

class ConvoMessage(BaseModel):
    history: str # it could be the whole original message or the summarization of the previous conversation from memory module
    message: str


class OrchestratorMessage(BaseModel):
    message: str
    attribute: dict


### Slot-related classes

class Slot(BaseModel):
    name: str
    type: Union[str, int, float, bool, None]
    value: Union[str, int, float, bool, List[str], None]
    enum: Optional[list[Union[str, int, float, bool, None]]]
    description: str
    prompt: str
    required: bool
    verified: bool


class Slots(BaseModel):
    slots: list[Slot]

# May not be needed
class SlotDetail(Slot):
    verified_value: str
    confirmed: bool

class Verification(BaseModel):
    thought: str
    verification_needed: bool


### Task status-related classes

class StatusEnum(Enum):
    COMPLETE = "complete"
    INCOMPLETE = "incomplete"


class MessageState(TypedDict):
    # system configuration
    sys_instruct: str
    # bot configuration
    bot_config: BotConfig
    # input message
    user_message: ConvoMessage
    orchestrator_message: OrchestratorMessage
    # action trajectory
    trajectory: list[dict[str, any]]
    # message flow between different nodes
    message_flow: Annotated[str, "message flow between different nodes"]
    # final response
    response: str
    # task-related params
    status: StatusEnum
    slots: dict[str, list[Slot]]  # record the dialogue states of each action
    metadata: dict[str, any]
    # stream
    is_stream: bool
    message_queue: janus.SyncQueue


from ..tools import register_tool, logger
import ast
import hubspot
from hubspot.crm.objects.emails import ApiException
from hubspot.crm.associations.v4 import AssociationSpec
from hubspot.crm.tickets.models import SimplePublicObjectInputForCreate
from arklex.env.tools.hubspot.utils import HUBSPOT_AUTH_ERROR
from datetime import datetime

description = "Create a ticket for the existing customer when the customer has some problem about the specific product."


slots = [
    {
        "name": "contact_information",
        "type": "string",
        "description": "After finding the exiting customer, the detailed information of the customer (actually a dict) is provided",
        "prompt": "",
        "required": True,
    },
    {
        "name": "issue",
        "type": "string",
        "description": "The question that the customer has for the specific product",
        "prompt": "",
        "required": True,
    }
]
outputs = [
    {
        "name": "ticket_information",
        "type": "string",
        "description": "The basic ticket information for the existing customer and the specific issue",
    }
]

USER_NOT_FOUND_ERROR = "error: user not found (not an existing customer)"
errors = [
    HUBSPOT_AUTH_ERROR
]


@register_tool(description, slots, outputs, lambda x: x not in errors)
def create_ticket(contact_information: str, issue: str, **kwargs) -> str:
    access_token = kwargs.get('access_token')
    contact_information = ast.literal_eval(contact_information)
    contact_id = contact_information.get('id')
    if not access_token:
        return HUBSPOT_AUTH_ERROR

    api_client = hubspot.Client.create(access_token=access_token)

    timestamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-3] + "Z"
    subject_name = "Issue of " + contact_id + " at " + timestamp
    ticket_properties = {
        'hs_pipeline_stage': 1,
        'content': issue,
        'subject': subject_name
    }
    ticket_for_create = SimplePublicObjectInputForCreate(properties=ticket_properties)
    try:
        ticket_creation_response = api_client.crm.tickets.basic_api.create(simple_public_object_input_for_create=ticket_for_create)
        ticket_creation_response = ticket_creation_response.to_dict()
        ticket_id = ticket_creation_response['id']
        association_spec = [
            AssociationSpec(
                association_category="HUBSPOT_DEFINED",
                association_type_id=15
            )
        ]
        ticket_information = {
            'id': ticket_id
        }
        try:
            association_creation_response = api_client.crm.associations.v4.basic_api.create(
                object_type="contact",
                object_id=contact_id,
                to_object_type="ticket",
                to_object_id=ticket_id,
                association_spec=association_spec
            )
            return str(ticket_information)
        except ApiException as e:
            logger.info("Exception when calling AssociationV4: %s\n" % e)
    except ApiException as e:
        logger.info("Exception when calling Crm.tickets.create: %s\n" % e)








import os
import logging
import uuid
import importlib
from typing import Optional

from arklex.env.tools.tools import Tool
from arklex.env.planner.function_calling import FunctionCallingPlanner
from arklex.utils.graph_state import StatusEnum
from arklex.orchestrator.NLU.nlu import SlotFilling


logger = logging.getLogger(__name__)

DEFAULT_WORKER  = {"id": "default_worker", "name": "DefaultWorker", "path": "default_worker.py"}

class BaseResourceInitializer:
    @staticmethod
    def init_tools(tools):
        raise NotImplementedError

    @staticmethod
    def init_workers(workers):
        raise NotImplementedError
    
class DefaulResourceInitializer(BaseResourceInitializer):
    @staticmethod
    def init_tools(tools):
        # return dict of valid tools with name and description
        tool_registry = {}
        for tool in tools:
            tool_id = tool["id"]
            name = tool["name"]
            path = tool["path"]
            try: # try to import the tool to check its existance
                filepath = os.path.join("arklex.env.tools", path)
                module_name = filepath.replace(os.sep, ".").rstrip(".py")
                module = importlib.import_module(module_name)
                func = getattr(module, name)
            except Exception as e:
                logger.error(f"Tool {name} is not registered, error: {e}")
                continue
            tool_registry[tool_id] = {
                "name": func().name,
                "description": func().description,
                "execute": func,
                "fixed_args": tool.get("fixed_args", {}),
            }
        return tool_registry
    
    @staticmethod
    def init_workers(workers):
        worker_registry = {}
        available_workers = [worker for worker in workers if worker["name"] != DEFAULT_WORKER["name"]]
        all_workers = [DEFAULT_WORKER] + available_workers if available_workers else [DEFAULT_WORKER]
        for worker in all_workers:
            worker_id = worker["id"]
            name = worker["name"]
            path = worker["path"]
            try: # try to import the worker to check its existance
                filepath = os.path.join("arklex.env.workers", path)
                module_name = filepath.replace(os.sep, ".").rstrip(".py")
                module = importlib.import_module(module_name)
                func = getattr(module, name)
            except Exception as e:
                logger.error(f"Worker {name} is not registered, error: {e}")
                continue
            worker_registry[worker_id] = {
                "name": name,
                "description": func().description,
                "execute": func
            }
        return worker_registry

class Env():
    def __init__(self, tools, workers, slotsfillapi, resource_inizializer: Optional[BaseResourceInitializer] = None):
        if resource_inizializer is None:
            resource_inizializer = DefaulResourceInitializer()
        self.tools = resource_inizializer.init_tools(tools)
        self.workers = resource_inizializer.init_workers(workers)
        self.name2id = {resource["name"]: id for id, resource in {**self.tools, **self.workers}.items()}
        self.id2name = {id: resource["name"] for id, resource in {**self.tools, **self.workers}.items()}
        self.slotfillapi = self.initialize_slotfillapi(slotsfillapi)
        self.planner = FunctionCallingPlanner(
            tools_map=self.tools,
            name2id=self.name2id
        )

    def initialize_slotfillapi(self, slotsfillapi):
        return SlotFilling(slotsfillapi)

    def step(self, id, message_state, params):
        if id in self.tools:
            logger.info(f"{self.tools[id]['name']} tool selected")
            tool: Tool = self.tools[id]["execute"]()
            tool.init_slotfilling(self.slotfillapi)
            response_state = tool.execute(message_state, **self.tools[id]["fixed_args"])
            params["history"] = response_state.get("trajectory", [])
            params["dialog_states"] = response_state.get("slots", [])
            current_node = params.get("curr_node")
            params["node_status"][current_node] = response_state.get("status", StatusEnum.COMPLETE.value)
                
        elif id in self.workers:
            message_state["metadata"]["worker"] = self.workers
            logger.info(f"{self.workers[id]['name']} worker selected")
            worker = self.workers[id]["execute"]()
            response_state = worker.execute(message_state)
            call_id = str(uuid.uuid4())
            params["history"].append({'content': None, 'role': 'assistant', 'tool_calls': [{'function': {'arguments': "{}", 'name': self.id2name[id]}, 'id': call_id, 'type': 'function'}], 'function_call': None})
            params["history"].append({
                        "role": "tool",
                        "tool_call_id": call_id,
                        "name": self.id2name[id],
                        "content": response_state.get("response", "")
            })
        else:
            logger.info("planner selected")
            action, response_state, msg_history = self.planner.execute(message_state, params["history"])
        
        logger.info(f"Response state from {id}: {response_state}")
        return response_state, params



import json
import logging

import shopify

# general GraphQL navigation utilities
from arklex.env.tools.shopify.utils_nav import *
from arklex.env.tools.shopify.utils import authorify_admin

# ADMIN
from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs
from arklex.env.tools.tools import register_tool

from arklex.utils.model_provider_config import PROVIDER_MAP
from arklex.utils.model_config import MODEL
from langchain_openai import ChatOpenAI


logger = logging.getLogger(__name__)

description = "Get the product image url of a product."
slots = [
    ShopifySlots.PRODUCT_IDS,
    *PAGEINFO_SLOTS
]
outputs = [
    ShopifyOutputs.PRODUCTS_DETAILS,
    *PAGEINFO_OUTPUTS
]
PRODUCTS_NOT_FOUND = "error: product not found"
errors = [PRODUCTS_NOT_FOUND]

@register_tool(description, slots, outputs, lambda x: x not in errors, True)
def get_product_images(product_ids: list, **kwargs) -> str:
    nav = cursorify(kwargs)
    if not nav[1]:
        return nav[0]
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]

    try:
        ids = ' OR '.join(f'id:{pid.split("/")[-1]}' for pid in product_ids)
        with shopify.Session.temp(**auth["value"]):
            response = shopify.GraphQL().execute(f"""
                {{
                    products ({nav[0]}, query:"{ids}") {{
                        nodes {{
                            title
                            description
                            onlineStoreUrl
                            images(first: 1) {{
                                edges {{
                                    node {{
                                        src
                                        altText
                                    }}
                                }}
                            }}
                        }}
                        pageInfo {{
                            endCursor
                            hasNextPage
                            hasPreviousPage
                            startCursor
                        }}
                    }}
                }}
            """)
            result = json.loads(response)['data']['products']
            response = result["nodes"]
            answer = "Here are images of products:\n"
            card_list = []
            for product in response:
                product_dict = {
                    "id": product.get('id'),
                    "title": product.get('title'), 
                    "description": product.get('description', "None")[:180] + "...", 
                    "link_url": product.get('onlineStoreUrl'),
                    "image_url" : product.get('images', {}).get('edges', [{}])[0].get('node', {}).get('src', ""),
                }
                card_list.append(product_dict)
            if card_list:
                try:
                    llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(model=MODEL["model_type_or_path"], timeout=30000)
                    message = [
                        {"role": "system", "content": f"You are helping a customer search products based on the query and get results below and those results will be presented using product card format.\n\n{json.dumps(card_list)}\n\nGenerate a response to continue the conversation without explicitly mentioning contents of the search result. Include one or two questions about those products to know the user's preference. Keep the response within 50 words.\nDIRECTLY GIVE THE RESPONSE."},
                    ]
                    answer = llm.invoke(message).content
                except Exception as e:
                    logger.info(f"llm error in search_products: {e}")
                    pass
                return json.dumps({
                    "answer": answer,
                    "card_list": card_list
                })
            else:
                return json.dumps({
                    "answer": PRODUCTS_NOT_FOUND,
                    "card_list": []
                })
    except Exception as e:
        return json.dumps({
                "answer": PRODUCTS_NOT_FOUND,
                "card_list": []
            })



import os
import json
import argparse
import time
import logging
import subprocess
import signal
import atexit
from dotenv import load_dotenv

from langchain_openai import ChatOpenAI

from arklex.utils.utils import init_logger
from arklex.orchestrator.orchestrator import AgentOrg
from arklex.orchestrator.generator.generator import Generator
from arklex.env.tools.RAG.build_rag import build_rag
from arklex.env.tools.database.build_database import build_database
from arklex.utils.model_config import MODEL

logger = init_logger(log_level=logging.INFO, filename=os.path.join(os.path.dirname(__file__), "logs", "arklex.log"))
load_dotenv()

# API_PORT = "55135"
# NLUAPI_ADDR = f"http://localhost:{API_PORT}/nlu"
# SLOTFILLAPI_ADDR = f"http://localhost:{API_PORT}/slotfill"

def generate_taskgraph(args):
    model = ChatOpenAI(model=MODEL["model_type_or_path"], timeout=30000)
    generator = Generator(args, args.config, model, args.output_dir)
    taskgraph_filepath = generator.generate()
    # Update the task graph with the API URLs
    task_graph = json.load(open(os.path.join(os.path.dirname(__file__), taskgraph_filepath)))
    task_graph["nluapi"] = ""
    task_graph["slotfillapi"] = ""
    with open(taskgraph_filepath, "w") as f:
        json.dump(task_graph, f, indent=4)


def init_worker(args):
    ## TODO: Need to customized based on different use cases
    config = json.load(open(args.config))
    workers = config["workers"]
    worker_names = set([worker["name"] for worker in workers])
    if "FaissRAGWorker" in worker_names:
        logger.info("Initializing FaissRAGWorker...")
        # if url: uncomment the following line
        build_rag(args.output_dir, config["rag_docs"])
        # if shopify: uncomment the following lines
        # import shopify
        # from arklex.utils.loaders.shopify import ShopifyLoader
        # session = shopify.Session(os.environ["SHOPIFY_SHOP_URL"], os.environ["SHOPIFY_API_VERSION"], os.environ["SHOPIFY_ACCESS_TOKEN"])
        # shopify.ShopifyResource.activate_session(session)
        # loader = ShopifyLoader()
        # docs = loader.load()
        # filepath = os.path.join(args.output_dir, "documents.pkl")
        # ShopifyLoader.save(filepath, docs)
        # chunked_docs = loader.chunk(docs)
        # filepath_chunk = os.path.join(args.output_dir, "chunked_documents.pkl")
        # ShopifyLoader.save(filepath_chunk, chunked_docs)
        

    elif any(node in worker_names for node in ("DataBaseWorker", "search_show", "book_show", "check_booking", "cancel_booking")):
        logger.info("Initializing DataBaseWorker...")
        build_database(args.output_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', type=str, default="./arklex/orchestrator/examples/customer_service_config.json")
    parser.add_argument('--output-dir', type=str, default="./examples/test")
    parser.add_argument('--model', type=str, default=MODEL["model_type_or_path"])
    parser.add_argument('--log-level', type=str, default="INFO", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    parser.add_argument('--task', type=str, choices=["gen_taskgraph", "init", "all"], default="all")
    args = parser.parse_args()
    MODEL["model_type_or_path"] = args.model
    log_level = getattr(logging, args.log_level.upper(), logging.INFO)
    logger = init_logger(log_level=log_level, filename=os.path.join(os.path.dirname(__file__), "logs", "arklex.log"))

    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir, exist_ok=True)
    
    if args.task == "all":
        generate_taskgraph(args)
        init_worker(args)
    elif args.task == "gen_taskgraph":
        generate_taskgraph(args)
    elif args.task == "init":
        init_worker(args)


import abc

from pydantic import BaseModel

from benchmark.tau_bench.model_utils.api.datapoint import Datapoint, ScoreDatapoint
from benchmark.tau_bench.model_utils.model.model import Model


class RequestRouter(abc.ABC):
    @abc.abstractmethod
    def route(self, dp: Datapoint, available_models: list[Model]) -> Model:
        raise NotImplementedError


class FirstModelRequestRouter(RequestRouter):
    def route(self, dp: Datapoint, available_models: list[Model]) -> Model:
        supporting_models = [model for model in available_models if model.supports_dp(dp)]
        if len(supporting_models) == 0:
            raise ValueError(f"No supporting models found from {available_models}")
        return supporting_models[0]


class CapabilityScoreModel(abc.ABC):
    @abc.abstractmethod
    def score_dp(self, dp: Datapoint) -> float:
        raise NotImplementedError


class PromptedLLMCapabilityScoreModel:
    def __init__(self, model: Model | None = None) -> None:
        if model is None:
            from benchmark.tau_bench.model_utils.model.claude import ClaudeModel

            # claude is used as the default model as it is better at meta-level tasks
            model = ClaudeModel()
        self.model = model

    def score_dp(self, dp: Datapoint, examples: list[ScoreDatapoint] | None = None) -> float:
        return (
            self.model.score(
                instruction="Score the task in the datapoint on a scale of 1 (least complex) to 10 (most complex).",
                text=f"----- start task -----\n{dp.model_dump_json()}\n----- end task -----",
                min=1,
                max=10,
                examples=examples,
            )
            / 10.0
        )


class MinimumCapabilityRequestRouter(RequestRouter):
    def __init__(self, capability_score_model: CapabilityScoreModel) -> None:
        self.capability_score_model = capability_score_model

    def route(self, dp: Datapoint, available_models: list[Model]) -> Model:
        supporting_models = [model for model in available_models if model.supports_dp(dp)]
        if len(supporting_models) == 0:
            raise ValueError(f"No supporting models found from {available_models}")
        required_capability = self.capability_score_model.score_dp(dp)
        minimum_model: Model | None = None
        minimum_model_capability: float | None = None
        for model in supporting_models:
            capability = model.get_capability()
            if capability >= required_capability and (
                minimum_model_capability is None or capability < minimum_model_capability
            ):
                minimum_model = model
                minimum_model_capability = capability
        if minimum_model is None:
            raise ValueError(f"No model found with capability >= {required_capability}")
        return minimum_model


def request_router_factory(
    router_id: str, capability_score_model: CapabilityScoreModel | None = None
) -> RequestRouter:
    if router_id == "first-model":
        return FirstModelRequestRouter()
    elif router_id == "minimum-capability":
        if capability_score_model is None:
            raise ValueError("CapabilityScoreModel is required for minimum-capability router")
        return MinimumCapabilityRequestRouter(capability_score_model=capability_score_model)
    raise ValueError(f"Unknown router_id: {router_id}")


def default_request_router() -> RequestRouter:
    return FirstModelRequestRouter()


class RequestRouteDatapoint(BaseModel):
    dp: Datapoint
    capability_score: float



PAGEINFO_SLOTS = [
    {
        "name": "limit",
        "type": "string",
        "description": "Maximum number of entries to show.",
        "prompt": "",
        "required": False
    },
    {
        "name": "navigate",
        "type": "string",
        "description": "navigate relative to previous view. 'next' to search after previous view, 'prev' to search before the previous view. 'stay' or None to remain.'",
        "prompt": "",
        "required": False
    },
    {
        "name": "pageInfo",
        "type": "string",
        "description": "The previous pageInfo object, such as \"{'endCursor': 'eyJsYXN0X2lkIjo3Mjk2NTgxODk0MjU3LCJsYXN0X3ZhbHVlIjoiNzI5NjU4MTg5NDI1NyJ9', 'hasNextPage': True, 'hasPreviousPage': False, 'startCursor': 'eyJsYXN0X2lkIjo3Mjk2NTgwODQ1NjgxLCJsYXN0X3ZhbHVlIjoiNzI5NjU4MDg0NTY4MSJ9'}\"",
        "prompt": "",
        "required": False
    },
]
    
PAGEINFO_OUTPUTS = [
    {
        "name": "pageInfo",
        "type": "string",
        "description": "Current pageInfo object, such as  \"{'endCursor': 'eyJsYXN0X2lkIjo3Mjk2NTgxODk0MjU3LCJsYXN0X3ZhbHVlIjoiNzI5NjU4MTg5NDI1NyJ9', 'hasNextPage': True, 'hasPreviousPage': False, 'startCursor': 'eyJsYXN0X2lkIjo3Mjk2NTgwODQ1NjgxLCJsYXN0X3ZhbHVlIjoiNzI5NjU4MDg0NTY4MSJ9'}\""
    }
]

NAVIGATE_WITH_NO_CURSOR = "error: cannot navigate without reference cursor"
NO_NEXT_PAGE = "error: no more pages after"
NO_PREV_PAGE = "error: no more pages before"
    
def cursorify(kwargs):
    limit = int(kwargs.get('limit')) if kwargs.get('limit') else 3
    navigate = kwargs.get('navigate') if kwargs.get('navigate') else 'stay'
    pageInfo = kwargs.get('pageInfo')
    
    nav_param = f'first: {limit}'
    if navigate and navigate != 'stay':
        if not pageInfo:
            return NAVIGATE_WITH_NO_CURSOR, False
        
        if navigate == 'next':
            if not pageInfo['hasNextPage']:
                return NO_NEXT_PAGE, False
            nav_param = f"first: {limit}, after: \"{pageInfo['endCursor']}\""
            
        elif navigate == 'prev': 
            if not pageInfo['hasPreviousPage']:
                return NO_PREV_PAGE, False
            nav_param = f"last: {limit}, before: \"{pageInfo['startCursor']}\""
    
    return nav_param, True